#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Dict, Optional, Union, Any, Mapping
import enum
from fastapi import Body, Request, HTTPException, Query as FastApiQuery
import pydantic
from pydantic import BaseModel, Field, validator, root_validator
import datetime
import numpy as np
import re
import fnmatch

from nomad import datamodel  # pylint: disable=unused-import
from nomad.utils import strip
from nomad.metainfo import Datetime, MEnum
from nomad.metainfo.search_extension import metrics, search_quantities, search_sub_sections

from .utils import parameter_dependency_from_model, update_url_query_arguments


User = datamodel.User.m_def.a_pydantic.model


calc_id = 'calc_id'
Metric = enum.Enum('Metric', {name: name for name in metrics})  # type: ignore
AggregateableQuantity = enum.Enum('AggregateableQuantity', {  # type: ignore
    name: name for name in search_quantities
    if search_quantities[name].aggregateable})

AggregateableQuantity.__doc__ = '''
    Statistics and aggregations can only be computed for those search quantities that have
    discrete values. For example a statistics aggregates a certain metric (e.g. the number of entries)
    over all entries were this quantity has the same value (bucket aggregation, think historgam here).
'''

Value = Union[str, int, float, bool, datetime.datetime]
ComparableValue = Union[str, int, float, datetime.datetime]


class AggregationOrderType(str, enum.Enum):
    '''
    Allows to order statistics or aggregations by either quantity values (`values`) or number
    of entries (`entries`).
    '''
    values = 'values'
    entries = 'entries'


class HTTPExceptionModel(BaseModel):
    detail: str


class NoneEmptyBaseModel(BaseModel):
    @root_validator
    def check_exists(cls, values):  # pylint: disable=no-self-argument
        assert any(value is not None for value in values.values())
        return values


class All(NoneEmptyBaseModel):
    op: List[Value] = Field(None, alias='all')


class None_(NoneEmptyBaseModel):
    op: List[Value] = Field(None, alias='none')


class Any_(NoneEmptyBaseModel):
    op: List[Value] = Field(None, alias='any')


class ComparisonOperator(NoneEmptyBaseModel): pass


class Lte(ComparisonOperator):
    op: ComparableValue = Field(None, alias='lte')


class Lt(ComparisonOperator):
    op: ComparableValue = Field(None, alias='lt')


class Gte(ComparisonOperator):
    op: ComparableValue = Field(None, alias='gte')


class Gt(ComparisonOperator):
    op: ComparableValue = Field(None, alias='gt')


class LogicalOperator(NoneEmptyBaseModel):

    @validator('op', check_fields=False)
    def validate_query(cls, query):  # pylint: disable=no-self-argument
        if isinstance(query, list):
            return [_validate_query(item) for item in query]

        return _validate_query(query)


class And(LogicalOperator):
    op: List['Query'] = Field(None, alias='and')


class Or(LogicalOperator):
    op: List['Query'] = Field(None, alias='or')


class Not(LogicalOperator):
    op: 'Query' = Field(None, alias='not')


ops = {
    'lte': Lte,
    'lt': Lt,
    'gte': Gte,
    'gt': Gt,
    'all': All,
    'none': None_,
    'any': Any_
}


QueryParameterValue = Union[Value, List[Value], Lte, Lt, Gte, Gt, Any_, All, None_]

Query = Union[
    Mapping[str, QueryParameterValue], And, Or, Not]


And.update_forward_refs()
Or.update_forward_refs()
Not.update_forward_refs()


class Owner(str, enum.Enum):
    '''
    The `owner` allows to limit the scope of the searched based on entry ownership.
    This is useful, if you only want to search among all publically downloadable
    entries, or only among your own entries, etc.

    These are the possible owner values and their meaning:
    * `all`: Consider all entries.
    * `public` (default): Consider all entries that can be publically downloaded,
        i.e. only published entries without embargo
    * `user`: Only consider entries that belong to you.
    * `shared`: Only consider entries that belong to you or are shared with you.
    * `visible`: Consider all entries that are visible to you. This includes
        entries with embargo or unpublished entries that belong to you or are
        shared with you.
    * `staging`: Only search through unpublished entries.
    '''

    # There seems to be a slight bug in fast API. When it creates the example in OpenAPI
    # it will ignore any given default or example and simply take the first enum value.
    # Therefore, we put public first, which is the most default and save in most contexts.
    public = 'public'
    all_ = 'all'
    visible = 'visible'
    shared = 'shared'
    user = 'user'
    staging = 'staging'
    admin = 'admin'


class WithQuery(BaseModel):
    owner: Optional[Owner] = Body('public')
    query: Optional[Query] = Body(
        None,
        embed=True,
        description=strip('''
            A query can be very simple list of parameters. Different parameters are combined
            with a logical **and**, values of the same parameter with also with a logical **and**.
            The following would search for all entries that are VASP calculations,
            contain *Na* **and** *Cl*, **and** are authored by *Stefano Curtarolo*
            **and** *Chris Wolverton*.
            ```
            {
                "atoms": ["Na", "Cl"],
                "dft.code_name": "VASP",
                "authors": ["Stefano Curtarolo", "Chris Wolverton"]
            }
            ```

            A short cut to change the logical combination of values in a list, is to
            add a suffix to the quantity `:any`:
            ```
            {
                "atoms": ["Na", "Cl"],
                "dft.code_name": "VASP",
                "authors:any": ["Stefano Curtarolo", "Chris Wolverton"]
            }
            ```

            Otherwise, you can also write complex logical combinations of parameters like this:
            ```
            {
                "and": [
                    {
                        "or": [
                            {
                                "atoms": ["Cl", "Na"]
                            },
                            {
                                "atoms": ["H", "O"]
                            }
                        ]
                    },
                    {
                        "not": {
                            "dft.crystal": "cubic"
                        }
                    }
                ]
            }
            ```
            Other short-cut prefixes are `none:` and `any:` (the default).

            By default all quantity values have to **equal** the given values to match. For
            some values you can also use comparison operators like this:
            ```
            {
                "upload_time": {
                    "gt": "2020-01-01",
                    "lt": "2020-08-01"
                },
                "dft.workflow.section_geometry_optimization.final_energy_difference": {
                    "lte": 1.23e-18
                }
            }
            ```

            or shorter with suffixes:
            ```
            {
                "upload_time:gt": "2020-01-01",
                "upload_time:lt": "2020-08-01",
                "dft.workflow.section_geometry_optimization.final_energy_difference:lte" 1.23e-18
            }
            ```

            The searchable quantities are a subset of the NOMAD Archive quantities defined
            in the NOMAD Metainfo. The most common quantities are: %s.
        ''' % ', '.join(reversed([
            '`%s`' % name
            for name in search_quantities
            if (name.startswith('dft') or '.' not in name) and len(name) < 20
        ]))),
        example={
            'upload_time:gt': '2020-01-01',
            'atoms': ['Ti', 'O'],
            'dft.code_name': 'VASP',
            'dft.workflow.section_geometry_optimization.final_energy_difference:lte': 1.23e-18,
            'dft.quantities': 'section_dos',
            'dft.system:any': ['bulk', '2d']
        })

    @validator('query')
    def validate_query(cls, query):  # pylint: disable=no-self-argument
        return _validate_query(query)


def _validate_query(query: Query):
    if isinstance(query, dict):
        for key, value in list(query.items()):
            # Note, we loop over a list of items, not query.items(). This is because we
            # may modify the query in the loop.
            if ':' in key:
                quantity, qualifier = key.split(':')
            else:
                quantity, qualifier = key, None

            assert quantity in search_quantities, '%s is not a searchable quantity' % key
            if qualifier is not None:
                assert quantity not in query, 'a quantity can only appear once in a query'
                assert qualifier in ops, 'unknown quantity qualifier %s' % qualifier
                del(query[key])
                query[quantity] = ops[qualifier](**{qualifier: value})  # type: ignore
            elif isinstance(value, list):
                query[quantity] = All(all=value)

    return query


def query_parameters(
    request: Request,
    owner: Optional[Owner] = FastApiQuery(
        'public', description=strip(Owner.__doc__)),
    q: Optional[List[str]] = FastApiQuery(
        [], description=strip('''
            Since we cannot properly offer forms for all parameters in the OpenAPI dashboard,
            you can use the parameter `q` and encode a query parameter like this
            `atoms__H` or `n_atoms__gt__3`. Multiple usage of `q` will combine parameters with
            logical *and*.
        '''))) -> WithQuery:

    # copy parameters from request
    query_params = {
        key: request.query_params.getlist(key)
        for key in request.query_params}

    # add the encoded parameters
    for parameter in q:
        fragments = parameter.split('__')
        if len(fragments) == 1 or len(fragments) > 3:
            raise HTTPException(422, detail=[{
                'loc': ['query', 'q'],
                'msg': 'wrong format, use <quantity>[__<op>]__<value>'}])
        name_op, value = '__'.join(fragments[:-1]), fragments[-1]
        quantity_name = name_op.split('__')[0]

        if quantity_name not in search_quantities:
            raise HTTPException(422, detail=[{
                'loc': ['query', parameter],
                'msg': '%s is not a search quantity' % quantity_name}])

        query_params.setdefault(name_op, []).append(value)

    # transform query parameters to query
    query: Dict[str, Any] = {}
    for key in query_params:
        op = None
        if '__' in key:
            quantity_name, op = key.split('__')
        else:
            quantity_name = key

        if quantity_name not in search_quantities:
            continue

        quantity = search_quantities[quantity_name]
        type_ = quantity.definition.type
        if type_ is Datetime:
            type_ = datetime.datetime.fromisoformat
        elif isinstance(type_, MEnum):
            type_ = str
        elif isinstance(type_, np.dtype):
            type_ = float
        elif type_ not in [int, float, bool]:
            type_ = str
        values = query_params[key]
        values = [type_(value) for value in values]

        if op is None:
            if quantity.many_and:
                op = 'all'
            if quantity.many_or:
                op = 'any'

        if op is None:
            if len(values) > 1:
                raise HTTPException(
                    status_code=422,
                    detail=[{
                        'loc': ['query', key],
                        'msg':'search parameter %s does not support multiple values' % key}])
            query[quantity_name] = values[0]

        elif op == 'all':
            query[quantity_name] = All(all=values)
        elif op == 'any':
            query[quantity_name] = Any_(any=values)
        elif op in ops:
            if len(values) > 1:
                raise HTTPException(
                    status_code=422,
                    detail=[{
                        'loc': ['query', key],
                        'msg': 'operator %s does not support multiple values' % op}])
            query[quantity_name] = ops[op](**{op: values[0]})
        else:
            raise HTTPException(
                422, detail=[{'loc': ['query', key], 'msg': 'operator %s is unknown' % op}])

    return WithQuery(query=query, owner=owner)


class Direction(str, enum.Enum):
    '''
    Order direction, either ascending (`asc`) or descending (`desc`)
    '''
    asc = 'asc'
    desc = 'desc'


class MetadataRequired(BaseModel):
    ''' Defines which metadata quantities are included or excluded in the response. '''

    include: Optional[List[str]] = Field(
        None, description=strip('''
            Quantities to include for each result. Only those quantities will be
            returned. The entry id quantity `calc_id` will always be included.
        '''))
    exclude: Optional[List[str]] = Field(
        None, description=strip('''
            Quantities to exclude for each result. Only all other quantities will
            be returned. The quantity `calc_id` cannot be excluded.
        '''))

    @validator('include', 'exclude')
    def validate_include(cls, value, values, field):  # pylint: disable=no-self-argument
        if value is None:
            return None

        for item in value:
            assert item in search_quantities or item in search_sub_sections or item[-1] == '*', \
                f'required fields ({item}) must be valid search quantities or contain wildcards'

        if field.name == 'include' and 'calc_id' not in value:
            value.append('calc_id')

        if field.name == 'exclude':
            if 'calc_id' in value:
                value.remove('calc_id')

        return value


metadata_required_parameters = parameter_dependency_from_model(
    'metadata_required_parameters', MetadataRequired)


class Pagination(BaseModel):
    ''' Defines the order, size, and page of results. '''

    page_size: Optional[int] = Field(
        10, description=strip('''
            The page size, e.g. the maximum number of items contained in one response.
            A `page_size` of 0 will return no results.
        '''))
    order_by: Optional[str] = Field(
        None,  # type: ignore
        description=strip('''
            The results are ordered by the values of this field. If omitted, default
            ordering is applied.
        '''))
    order: Optional[Direction] = Field(
        Direction.asc, description=strip('''
            The ordering direction of the results based on `order_by`. Its either
            ascending `asc` or decending `desc`. Default is `asc`.
        '''))
    page_after_value: Optional[str] = Field(
        None, description=strip('''
            This attribute defines the position after which the page begins, and is used
            to navigate through the total list of results.

            When requesting the first page, no value should be provided for
            `page_after_value`. Each response will contain a value `next_page_after_value`,
            which can be used to obtain the next page (by setting `page_after_value` in
            your next request to this value).

            The field is encoded as a string, and the format of `page_after_value` and
            `next_page_after_value` depends on which API method is used.

            Some API functions additionally allows a simplified navigation, by specifying
            the page number in the key `page`. It is however always possible to use
            `page_after_value` and `next_page_after_value` to iterate through the results.
            '''))
    page: Optional[int] = Field(
        None, description=strip('''
            The number of the page (1-based). When provided in a request, this attribute
            can be used instead of `page_after_value` to jump to a particular results page.

            **NOTE #1**: the option to request pages by submitting the `page` number is
            limited. There are api calls where this attribute cannot be used for indexing,
            or where it can only be used partially. **If you want to just iterate through
            all the results, aways use the `page_after_value` and `next_page_after_value`!**

            **NOTE #2**: In a request, you should never specify *both* `page` and
            `page_after_value` (at most one index can be provided).
        '''))

    @validator('page_size')
    def validate_page_size(cls, page_size):  # pylint: disable=no-self-argument
        assert page_size >= 0, 'page_size must be >= 0'
        return page_size

    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        '''
        Override this in your Pagination class to ensure that a valid attribute is selected.
        This method has to be implemented!
        '''
        raise NotImplementedError('Validation of `order_by` not implemented!')

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        '''
        Override this in your Pagination class to implement validation of the
        `page_after_value` value.
        This method has to be implemented!
        '''
        raise NotImplementedError('Validation of `page_after_value` not implemented!')

    @validator('page')
    def validate_page(cls, page, values):  # pylint: disable=no-self-argument
        if page is not None:
            assert page >= 1, 'page must be >= 1'
        return page

    @root_validator(skip_on_failure=True)
    def validate_values(cls, values):  # pylint: disable=no-self-argument
        # Because of a bug in pydantic (#2670), root validators can't be overridden, so
        # we invoke a class method, which *can* be overridden.
        return cls._root_validation(values)

    @classmethod
    def _root_validation(cls, values):
        page = values.get('page')
        page_after_value = values.get('page_after_value')
        page_size = values.get('page_size')
        assert page is None or page_after_value is None, 'Cannot specify both `page` and `page_after_value'
        if page_size == 0:
            assert page is None, 'Cannot specify `page` when `page_size` is set to 0'
            assert page_after_value is None, 'Cannot specify `page_after_value` when `page_size` is set to 0'
        return values

    def get_simple_index(self):
        '''
        If simple, index-based pagination is used, this method can be used to get the
        corresponding index (0-based). It will look on either `page` or `page_after_value`.
        If neither index is provided, we return 0 (i.e. the first index).
        '''
        if self.page is None and self.page_after_value is None:
            return 0
        if self.page is not None:
            return (self.page - 1) * self.page_size
        rv = int(self.page_after_value) + 1
        assert rv >= 0
        return rv


class PaginationResponse(Pagination):
    total: int = Field(
        ..., description=strip('''
        The total number of results that fit the given query. This is independent of
        any pagination and aggregations.
        '''))
    next_page_after_value: Optional[str] = Field(
        None, description=strip('''
        The *next* value to be used as `page_after_value` in a follow up requests, to get
        the next page of results. If no more results are available, `next_page_after_value`
        will not be set.
        '''))
    page_url: Optional[str] = Field(
        None, description=strip('''
        The url of the current page. Only applicable for GET requests.
        '''))
    next_page_url: Optional[str] = Field(
        None, description=strip('''
        The url to get the next page. Only applicable for GET requests.
        '''))
    prev_page_url: Optional[str] = Field(
        None, description=strip('''
        The url to get the previous page. **NOTE:** Only applicable for some API methods,
        (namely, where indexing by `page` is possible), and only for GET requests.
        '''))
    first_page_url: Optional[str] = Field(
        None, description=strip('''
        The url to get the first page. Only applicable for GET requests.
        '''))

    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        # No validation - behaviour of this field depends on api method
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # No validation - behaviour of this field depends on api method
        return page_after_value

    @classmethod
    def _root_validation(cls, values):  # pylint: disable=no-self-argument
        # No validation
        return values

    def populate_urls(self, request: Request):
        '''
        Populates the urls (`page_url`, `next_page_url`, `first_page_url` from the
        request and `next_page_after_value`. Only applicable for GET requests.
        '''
        assert request.method.upper() == 'GET', 'Trying to populate urls, but method is not GET.'
        original_url = str(request.url)
        self.page_url = original_url
        if self.page_size:
            self.first_page_url = update_url_query_arguments(
                original_url, page=None, page_after_value=None)
        if self.next_page_after_value:
            self.next_page_url = update_url_query_arguments(
                original_url, page=None, page_after_value=self.next_page_after_value)
        if self.page and self.page > 1:
            self.prev_page_url = update_url_query_arguments(
                original_url, page=self.page - 1, page_after_value=None)

    def populate_simple_index_and_urls(self, request: Request):
        '''
        If simple, index-based pagination is used, this method can be used to populate
        the `page`, `page_after_value` and urls (if it is a GET request) automatically.
        Assumes that the field `total` is populated.
        '''
        if not self.page_size:
            self.page = 1
            self.page_after_value = None
            self.next_page_after_value = None
        else:
            ind = self.get_simple_index()
            self.page = ind // self.page_size + 1
            self.page_after_value = None if self.page == 1 else str(ind - 1)
            if self.page_size * self.page >= self.total:
                self.next_page_after_value = None
            else:
                self.next_page_after_value = str(ind + self.page_size - 1)

            if self.page < 1 or (
                    self.total == 0 and self.page != 1) or (
                    self.total > 0 and (self.page - 1) * self.page_size >= self.total):
                raise HTTPException(400, detail='Page out of range requested.')
        if request.method.upper() == 'GET':
            self.populate_urls(request)


class EntryBasedPagination(Pagination):
    order_by: Optional[str] = Field(
        calc_id,  # type: ignore
        description=strip('''
            The results are ordered by the values of this field. If omitted, default
            ordering is applied.
        '''))

    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        if order_by is None:
            return order_by

        assert order_by in search_quantities, 'order_by must be a valid search quantity'
        quantity = search_quantities[order_by]
        assert quantity.definition.is_scalar, 'the order_by quantity must be a scalar'
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        order_by = values.get('order_by', calc_id)
        if page_after_value is not None and order_by is not None and order_by != calc_id:
            if ':' not in page_after_value:
                page_after_value = '%s:' % page_after_value
        return page_after_value


class EntryPagination(EntryBasedPagination):
    page: Optional[int] = Field(
        None, description=strip('''
            For simple, index-based pagination, this should contain the number of the
            requested page (1-based). When provided in a request, this attribute can be
            used instead of `page_after_value` to jump to a particular results page.

            However, you can only retreive up to the 10.000th entry with a page number.
            Only one, `page_after_value` *or* `page` can be provided.
        '''))

    @validator('page')
    def validate_page(cls, page, values):  # pylint: disable=no-self-argument
        if page is not None:
            assert page > 0, 'Page has to be larger than 1.'
            assert page * values.get('page_size', 10) < 10000, 'Pagination by `page` is limited to 10.000 entries.'

        return page


entry_pagination_parameters = parameter_dependency_from_model(
    'entry_pagination_parameters', EntryPagination)


class AggregationPagination(EntryBasedPagination):
    order_by: Optional[str] = Field(
        None,  # type: ignore
        description=strip('''
            The results are ordered by the values of this field. If omitted, default
            ordering is applied.
        '''))

    @validator('page')
    def validate_page(cls, page, values):  # pylint: disable=no-self-argument
        assert page is None, 'Pagination by `page` is not possible for aggregations, use `page_after_value`'
        return page


class AggregatedEntities(BaseModel):
    size: Optional[pydantic.conint(gt=0)] = Field(  # type: ignore
        1, description=strip('''
        The maximum number of entries that should be returned for each value in the
        aggregation.
        '''))
    required: Optional[MetadataRequired] = Field(
        None, description=strip('''
        This allows to determined what fields should be returned for each entry.
        '''))


class Aggregation(BaseModel):
    quantity: AggregateableQuantity = Field(
        ..., description=strip('''
        The manatory name of the quantity for the aggregation. Aggregations
        can only be computed for those search metadata that have discrete values;
        an aggregation buckets entries that have the same value for this quantity.'''))
    pagination: Optional[AggregationPagination] = Field(
        AggregationPagination(), description=strip('''
        Only the data few values are returned for each API call. Pagination allows to
        get the next set of values based on the last value in the last call.
        '''))
    entries: Optional[AggregatedEntities] = Field(
        None, description=strip('''
        Optionally, a set of entries can be returned for each value.
        '''))


class StatisticsOrder(BaseModel):
    type_: Optional[AggregationOrderType] = Field(AggregationOrderType.entries, alias='type')
    direction: Optional[Direction] = Field(Direction.desc)


class Statistic(BaseModel):
    quantity: AggregateableQuantity = Field(
        ..., description=strip('''
        The manatory name of the quantity that the statistic is calculated for. Statistics
        can only be computed for those search metadata that have discrete values; a statistics
        aggregates a certain metric (e.g. the number of entries) over all entries were
        this quantity has the same value (bucket aggregation, think historgam here).

        There is one except and these are date/time values quantities (most notably `upload_time`).
        Here each statistic value represents an time interval. The interval can
        be determined via `datetime_interval`.'''))
    metrics: Optional[List[Metric]] = Field(
        [], description=strip('''
        By default the returned statistics will provide the number of entries for each
        value. You can add more metrics. For each metric an additional number will be
        provided for each value. Metrics are also based on search metadata. Depending on
        the metric the number will represent either a sum (`calculations` for the number
        of individual calculation in each code run) or an amount of different values
        (i.e. `materials` for the amount of different material hashes).'''))
    datetime_interval: Optional[pydantic.conint(gt=0)] = Field(  # type: ignore
        None, description=strip('''
        While statistics in general are only possible for quantities with discrete values,
        these is one exception. These are date/time values quantities (most notably `upload_time`).
        Here each statistic value represents an time interval.

        A date/time interval is a number of seconds greater than 0. This will only be used for
        date/time valued quantities (e.g. `upload_time`).
        '''))
    value_filter: Optional[pydantic.constr(regex=r'^[a-zA-Z0-9_\-\s]+$')] = Field(  # type: ignore
        None, description=strip('''
        An optional filter for values. Only values that contain the filter as substring
        will be part of the statistics.
        '''))
    size: Optional[pydantic.conint(gt=0)] = Field(  # type: ignore
        None, description=strip('''
        An optional maximum size of values in the statistics. The default depends on the
        quantity.
        '''))
    order: Optional[StatisticsOrder] = Field(
        StatisticsOrder(), description=strip('''
        The values in the statistics are either ordered by the entry count or by the
        natural ordering of the values.
        '''))

    @root_validator(skip_on_failure=True)
    def fill_default_size(cls, values):  # pylint: disable=no-self-argument
        if 'size' not in values or values['size'] is None:
            values['size'] = search_quantities[values['quantity'].value].statistic_size

        return values


class WithQueryAndPagination(WithQuery):
    pagination: Optional[EntryPagination] = Body(
        None,
        example={
            'page_size': 5,
            'order_by': 'upload_time'
        })


class EntriesMetadata(WithQueryAndPagination):
    required: Optional[MetadataRequired] = Body(
        None,
        example={
            'include': ['calc_id', 'mainfile', 'upload_id', 'authors', 'upload_time']
        })
    statistics: Optional[Dict[str, Statistic]] = Body(
        {},
        description=strip('''
            This allows to define additional statistics that should be returned.
            Statistics aggregate entries that show the same quantity values for a given quantity.
            A simple example is the number of entries for each `dft.code_name`. These statistics
            will be computed only over the query results. This allows to get an overview about
            query results.
        '''),
        example={
            'by_code_name': {
                'metrics': ['uploads', 'datasets'],
                'quantity': 'dft.code_name'
            }
        })
    aggregations: Optional[Dict[str, Aggregation]] = Body(
        {},
        example={
            'uploads': {
                'quantity': 'upload_id',
                'pagination': {
                    'page_size': 10,
                    'order_by': 'upload_time'
                },
                'entries': {
                    'size': 1,
                    'required': {
                        'include': ['mainfile']
                    }
                }
            }
        },
        description=strip('''
            Defines additional aggregations to return. An aggregation list entries
            for the values of a quantity, e.g. to get all uploads and their entries.
        '''))


class Files(BaseModel):
    ''' Configures the download of files. '''
    compress: Optional[bool] = Field(
        False, description=strip('''
        By default the returned zip file is not compressed. This allows to enable compression.
        Compression will reduce the rate at which data is provided, often below
        the rate of the compression. Therefore, compression is only sensible if the
        network connection is limited.'''))
    glob_pattern: Optional[str] = Field(
        None, description=strip('''
        An optional *glob* (or unix style path) pattern that is used to filter the
        returned files. Only files matching the pattern are returned. The pattern is only
        applied to the end of the full path. Internally
        [fnmatch](https://docs.python.org/3/library/fnmatch.html) is used.'''))
    re_pattern: Optional[str] = Field(
        None, description=strip('''
        An optional regexp that is used to filter the returned files. Only files matching
        the pattern are returned. The pattern is applied in search mode to the full
        path of the files. With `$` and `^` you can control if you want to match the
        whole path.

        A re pattern will replace a given glob pattern.'''))

    @validator('glob_pattern')
    def validate_glob_pattern(cls, glob_pattern):  # pylint: disable=no-self-argument
        # compile the glob pattern into re
        if glob_pattern is None:
            return None

        return re.compile(fnmatch.translate(glob_pattern) + r'$')

    @validator('re_pattern')
    def validate_re_pattern(cls, re_pattern):  # pylint: disable=no-self-argument
        # compile an re
        if re_pattern is None:
            return None
        try:
            return re.compile(re_pattern)
        except re.error as e:
            assert False, 'could not parse the re pattern: %s' % e

    @root_validator()
    def vaildate(cls, values):  # pylint: disable=no-self-argument
        # use the compiled glob pattern as re
        if values.get('re_pattern') is None:
            values['re_pattern'] = values.get('glob_pattern')
        return values


files_parameters = parameter_dependency_from_model(
    'files_parameters', Files)


ArchiveRequired = Union[str, Dict[str, Any]]


_archive_required_field = Body(
    '*',
    embed=True,
    description=strip('''
        The `required` part allows you to specify what parts of the requested archives
        should be returned. The NOMAD Archive is a hierarchical data format and
        you can *require* certain branches (i.e. *sections*) in the hierarchy.
        By specifing certain sections with specific contents or all contents (via
        the directive `"*"`), you can determine what sections and what quantities should
        be returned. The default is the whole archive, i.e., `"*"`.

        For example to specify that you are only interested in the `section_metadata`
        use:

        ```
        {
            "section_metadata": "*"
        }
        ```

        Or to only get the `energy_total` from each individual calculations, use:
        ```
        {
            "section_run": {
                "section_single_configuration_calculation": {
                    "energy_total": "*"
                }
            }
        }
        ```

        You can also request certain parts of a list, e.g. the last calculation:
        ```
        {
            "section_run": {
                "section_single_configuration_calculation[-1]": "*"
            }
        }
        ```

        These required specifications are also very useful to get workflow results.
        This works because we can use references (e.g. workflow to final result calculation)
        and the API will resolve these references and return the respective data.
        For example just the total energy value and reduced formula from the resulting
        calculation:
        ```
        {
            "section_workflow": {
                "calculation_result_ref": {
                    "energy_total": "*",
                    "single_configuration_calculation_to_system_ref": {
                        "chemical_composition_reduced": "*"
                    }
                }
            }
        }
        ```

        You can also resolve all references in a branch with the `include-resolved`
        directive. This will resolve all references in the branch, and also all references
        in referenced sections:
        ```
        {
            "section_workflow":
                "calculation_result_ref": "include-resolved"
            }
        }
        ```

        By default, the targets of "resolved" references are added to the archive at
        their original hierarchy positions.
        This means, all references are still references, but they are resolvable within
        the returned data, since they targets are now part of the data. Another option
        is to add
        `"resolve-inplace": true` to the root of required. Here, the reference targets will
        replace the references:
        ```
        {
            "resolve-inplace": true,
            "section_workflow":
                "calculation_result_ref": "include-resolved"
            }
        }
        ```
    '''),
    example={
        'section_run': {
            'section_single_configuration_calculation[-1]': {
                'energy_total': '*'
            },
            'section_system[-1]': '*'
        },
        'section_metadata': '*'
    })


class EntriesArchive(WithQueryAndPagination):
    required: Optional[ArchiveRequired] = _archive_required_field


class EntryArchiveRequest(BaseModel):
    required: Optional[ArchiveRequired] = _archive_required_field


class EntriesArchiveDownload(WithQuery):
    files: Optional[Files] = Body(None)


class EntriesRaw(WithQuery):
    pagination: Optional[EntryPagination] = Body(None)


class EntriesRawDownload(WithQuery):
    files: Optional[Files] = Body(
        None,
        example={
            'glob_pattern': 'vasp*.xml*'
        })


class StatisticResponse(Statistic):
    data: Dict[str, Dict[str, int]] = Field(
        None, description=strip('''
        The returned statistics data as dictionary. The key is a string representation of the values.
        The concrete type depends on the quantity that was used to create the statistics.
        Each dictionary value is a dictionary itself. The keys are the metric names the
        values the metric values. The key `entries` that gives the amount of entries with
        this value is always returned.'''))


class AggregationDataItem(BaseModel):
    data: Optional[List[Dict[str, Any]]] = Field(
        None, description=strip('''The entries that were requested for each value.'''))
    size: int = Field(
        None, description=strip('''The amount of entries with this value.'''))


class AggregationResponse(Aggregation):
    pagination: PaginationResponse  # type: ignore
    data: Dict[str, AggregationDataItem] = Field(
        None, description=strip('''
        The aggregation data as a dictionary. The key is a string representation of the values.
        The dictionary values contain the aggregated data depending if `entries` where
        requested.'''))


class CodeResponse(BaseModel):
    curl: str
    requests: str
    nomad_lab: Optional[str]


class EntriesMetadataResponse(EntriesMetadata):
    pagination: PaginationResponse  # type: ignore
    statistics: Optional[Dict[str, StatisticResponse]]  # type: ignore
    aggregations: Optional[Dict[str, AggregationResponse]]  # type: ignore
    data: List[Dict[str, Any]] = Field(
        None, description=strip('''
        The entries data as a list. Each item is a dictionary with the metadata for each
        entry.'''))
    code: Optional[CodeResponse]


class EntryRawFile(BaseModel):
    path: str = Field(None)
    size: int = Field(None)


class EntryRaw(BaseModel):
    calc_id: str = Field(None)
    upload_id: str = Field(None)
    mainfile: str = Field(None)
    files: List[EntryRawFile] = Field(None)


class EntriesRawResponse(EntriesRaw):
    pagination: PaginationResponse = Field(None)  # type: ignore
    data: List[EntryRaw] = Field(None)


class EntryMetadataResponse(BaseModel):
    entry_id: str = Field(None)
    required: MetadataRequired = Field(None)
    data: Dict[str, Any] = Field(
        None, description=strip('''A dictionary with the metadata of the requested entry.'''))


class EntryRawResponse(BaseModel):
    entry_id: str = Field(...)
    data: EntryRaw = Field(...)


class EntryArchive(BaseModel):
    calc_id: str = Field(None)
    upload_id: str = Field(None)
    parser_name: str = Field(None)
    archive: Dict[str, Any] = Field(None)


class EntriesArchiveResponse(EntriesArchive):
    pagination: PaginationResponse = Field(None)  # type: ignore
    data: List[EntryArchive] = Field(None)


class EntryArchiveResponse(EntryArchiveRequest):
    entry_id: str = Field(...)
    data: EntryArchive = Field(None)


class SearchResponse(EntriesMetadataResponse):
    es_query: Any = Field(
        None, description=strip('''The elasticsearch query that was used to retrieve the results.'''))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Experimental material science specific metadata
'''
from nomad import config
from nomad.metainfo import Quantity, MSection, Section, Datetime
from nomad.metainfo.search_extension import Search


def _unavailable(value):
    if value is None:
        return config.services.unavailable_value

    return value


class EMSMetadata(MSection):
    m_def = Section(a_domain='ems')

    # sample quantities
    chemical = Quantity(type=str, a_search=Search())
    sample_constituents = Quantity(type=str, a_search=Search())
    sample_microstructure = Quantity(type=str, a_search=Search())

    # general metadata
    experiment_summary = Quantity(type=str, a_search=Search())
    origin_time = Quantity(type=Datetime, a_search=Search())
    experiment_location = Quantity(type=str, a_search=Search())

    # method
    method = Quantity(type=str, a_search=Search())
    data_type = Quantity(type=str, a_search=Search())
    probing_method = Quantity(type=str, a_search=Search())

    # data metadata
    repository_name = Quantity(type=str, a_search=Search())
    repository_url = Quantity(type=str, a_search=Search())
    entry_repository_url = Quantity(type=str, a_search=Search())
    preview_url = Quantity(type=str, a_search=Search())

    # TODO move
    quantities = Quantity(type=str, shape=['0..*'], default=[], a_search=Search())
    group_hash = Quantity(type=str, a_search=Search())

    def apply_domain_metadata(self, entry_archive):
        from nomad import utils

        if entry_archive is None:
            return

        entry = self.m_parent

        root_section = entry_archive.section_experiment
        entry.formula = root_section.section_sample.section_material.chemical_formula
        atoms = root_section.section_sample.section_material.atom_labels

        if atoms is None:
            entry.atoms = []
        else:
            if hasattr(atoms, 'tolist'):
                atoms = atoms.tolist()
            entry.n_atoms = len(atoms)

            atoms = list(set(atoms))
            atoms.sort()
            entry.atoms = atoms

        self.chemical = _unavailable(root_section.section_sample.section_material.chemical_name)
        self.sample_microstructure = _unavailable(root_section.section_sample.sample_microstructure)
        self.sample_constituents = _unavailable(root_section.section_sample.sample_constituents)

        self.experiment_summary = root_section.experiment_summary
        location = root_section.experiment_location
        if location is not None:
            location_str = ', '.join([
                getattr(location, prop)
                for prop in ['facility', 'institution', 'address']
                if getattr(location, prop) is not None])
            self.experiment_location = location_str

        if root_section.experiment_time:
            self.origin_time = root_section.experiment_time
        elif root_section.experiment_publish_time:
            self.origin_time = root_section.experiment_publish_time
        else:
            self.origin_time = self.m_parent.upload_time

        self.data_type = _unavailable(root_section.section_method.data_type)
        self.method = _unavailable(root_section.section_method.method_name)
        self.probing_method = _unavailable(root_section.section_method.probing_method)

        self.repository_name = _unavailable(root_section.section_data.repository_name)
        self.repository_url = root_section.section_data.repository_url
        self.preview_url = root_section.section_data.preview_url
        self.entry_repository_url = root_section.section_data.entry_repository_url

        self.group_hash = utils.hash(
            entry.formula,
            self.method,
            self.experiment_location,
            entry.with_embargo,
            entry.uploader)

        quantities = set()

        quantities.add(root_section.m_def.name)
        for _, property_def, _ in root_section.m_traverse():
            quantities.add(property_def.name)

        self.quantities = list(quantities)

        if self.m_parent.references is None:
            self.m_parent.references = [self.entry_repository_url]



import numpy as np
import logging
import ase
from ase.io.ulm import Reader

from .metainfo import m_env
from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser import FileParser, TarParser, XMLParser, DataTextParser
from nomad.datamodel.metainfo.common_dft import Run, BasisSetCellDependent, System,\
    BasisSetAtomCentered, SamplingMethod, Method, XCFunctionals, Eigenvalues,\
    SingleConfigurationCalculation, VolumetricData, KBand, KBandSegment


class GPWParser(TarParser):
    def __init__(self):
        super().__init__()
        self._version_map = {6: '1.1.0', 5: '0.11.0', 3: '0.10.0'}
        self._info_map = {
            'energy_total': 'Epot', 'energy_XC': 'Exc', 'electronic_kinetic_energy': 'Ekin',
            'energy_correction_entropy': 'S', 'atom_forces_free': 'CartesianForce',
            'atom_positions': 'CartesianPositions', 'occupation': 'OccupationNumbers',
            'kpoints': 'IBZKPoints'}

    def init_parameters(self):
        self._info = None

    @property
    def info(self):
        if self._info is None:
            self._info = dict()
            xml_file = self.get('info.xml', None)
            if xml_file is None:
                return
            self.xml_file = xml_file
            xml_parser = XMLParser(xml_file, logger=self.logger)

            def convert(val):
                if isinstance(val, list):
                    return [convert(v) for v in val]
                try:
                    if val in ['True', 'False']:
                        return val == 'True'
                    else:
                        val = float(val)
                        if val % 1.0 == 0.0:
                            val = int(val)
                        return val
                except Exception:
                    return val

            # parameters
            self._info['parameter'] = {
                'lengthunit': 'angstrom', 'energyunit': 'eV', 'timeunit': 'femtosecond'}
            self._info['parameter'].update({p['name'].lower(): convert(
                p['value']) for p in xml_parser.get('parameter/', [])})

            # array shapes, types, dimensions
            self._info['array'] = dict()
            dimension = dict()
            for arr in xml_parser.root.findall('./array'):
                name = arr.attrib.get('name', None)
                dtype = arr.attrib.get('type', None)
                if name is None or dtype is None:
                    continue
                shape = []
                for dim in arr.findall('./dimension'):
                    length = int(dim.attrib.get('length', 0))
                    shape.append(length)
                    dimension[dim.attrib.get('name')] = length
                self._info['array'][name.lower()] = dict(dtype=dtype, shape=shape)
            self._info['array_dimension'] = dimension

            self._info['bytesswap'] = (
                xml_parser.root.attrib['endianness'] == 'little') != np.little_endian
        return self._info

    def get_parameter(self, key, unit=None):
        key = self._info_map.get(key, key)
        return self.info['parameter'].get(key.lower(), None)

    def get_array(self, key, unit=None):
        key = self._info_map.get(key, key)
        file_object = self.get(key)
        if file_object is None:
            return

        key = key.lower()
        shape = self.info['array'].get(key, {}).get('shape', None)
        dtype = self.info['array'].get(key, {}).get('dtype', None)
        dtype = np.dtype({'int': 'int32'}.get(dtype, dtype))
        size = np.prod(shape) * dtype.itemsize

        file_object.seek(0)
        parser = DataTextParser(
            mainfile_contents=file_object.read(size), logger=self.logger, dtype=dtype)
        if parser.data is None:
            return

        array = parser.data
        if self._info['bytesswap']:
            array = array.byteswap()
        if dtype == np.int32:
            array = np.asarray(array, int)
        array.shape = shape
        return array

    def get_array_dimension(self, key):
        if key == 'ngpts':
            val = [self.get_array_dimension('ngpts%s' % n) for n in ['x', 'y', 'z']]
        else:
            val = self.info['array_dimension'].get(key)
        return val

    def get_program_version(self):
        return self._version_map.get(self.get_parameter('version'), '0.9.0')

    def get_smearing_width(self):
        return self.get_parameter('fermiwidth')


class GPW2Parser(FileParser):
    def __init__(self):
        super().__init__(None)

    def init_parameters(self):
        self._info = None

    @property
    def ulm(self):
        if self._file_handler is None:
            try:
                self._file_handler = Reader(self.mainfile)
            except Exception:
                pass
        return self._file_handler

    @property
    def info(self):
        if self._info is None:
            self._info = dict()
            self._info['parameter'] = {
                'mode': 'fd', 'xc': 'LDA', 'occupations': None, 'poissonsolver': None,
                'h': None, 'gpts': None, 'kpts': [(0.0, 0.0, 0.0)], 'nbands': None,
                'charge': 0, 'setups': {}, 'basis': {}, 'spinpol': None, 'fixdensity': False,
                'filter': None, 'mixer': None, 'eigensolver': None, 'background_charge': None,
                'external': None, 'random': False, 'hund': False, 'maxiter': 333,
                'idiotproof': True, 'symmetry': {
                    'point_group': True, 'time_reversal': True, 'symmorphic': True,
                    'tolerance': 1e-7},
                'convergence': {
                    'energy': 0.0005, 'density': 1.0e-4, 'eigenstates': 4.0e-8,
                    'bands': 'occupied', 'forces': np.inf},
                'dtype': None, 'width': None, 'verbose': 0, 'lengthunit': 'angstrom',
                'energyunit': 'eV', 'timeunit': 'femtosecond'}
            if self.ulm is not None:
                self._info['parameter'].update(self.ulm.parameters.asdict())
        return self._info

    def get_parameter(self, key):
        try:
            if key == 'planewavecutoff':
                val = self.get_parameter('mode').get('ecut', None)
            elif key == 'basisset':
                val = self.get_parameter('basis')
            elif key == 'energyerror':
                val = self.get_parameter('convergence').get('energy', None)
            elif key == 'xcfunctional':
                val = self.get_parameter('xc')
            # in gpw format, energies are parameters
            elif key == 'energy_total':
                val = self.ulm.hamiltonian.e_total_extrapolated
            elif key == 'energy_free':
                val = self.ulm.hamiltonian.e_total_free
            elif key == 'energy_XC':
                val = self.ulm.hamiltonian.e_xc
            elif key == 'electronic_kinetic_energy':
                val = self.ulm.hamiltonian.e_kinetic
            elif key == 'energy_correction_entropy':
                val = self.ulm.hamiltonian.e_entropy
            elif key == 'fermilevel':
                val = self.ulm.occupations.fermilevel
            elif key == 'split':
                val = self.ulm.occupations.split
            elif key == 'converged':
                val = self.ulm.scf.converged
            else:
                val = self.info['parameter'].get(key.lower(), None)
        except Exception:
            val = None
        return val

    def get_array(self, key):
        if self.ulm is None:
            return
        try:
            if key == 'unitcell':
                val = self.ulm.atoms.cell
            elif key == 'atomicnumbers':
                val = self.ulm.atoms.numbers
            elif key == 'atom_positions':
                val = self.ulm.atoms.positions
            elif key == 'boundaryconditions':
                val = self.ulm.atoms.pbc
            elif key == 'momenta':
                val = self.ulm.atoms.momenta
            elif key == 'atom_forces_free_raw':
                val = self.ulm.results.forces
            elif key == 'magneticmoments':
                val = self.ulm.results.magmoms
            elif key == 'eigenvalues':
                val = self.ulm.wave_functions.eigenvalues
            elif key == 'occupation':
                val = self.ulm.wave_functions.occupations
            # TODO no koints data in ulm?
            elif key == 'kpoints':
                val = self.ulm.IBZKPoints
            elif key == 'density':
                val = self.ulm.density.density
            elif key == 'potential_effective':
                val = self.ulm.hamiltonian.potential
            elif key == 'band_paths':
                val = self.ulm.wave_functions.band_paths.asdict()
            else:
                val = self.ulm.asdict().get(key, None)
        except Exception:
            val = None
        return val

    def get_array_dimension(self, key):
        if key == 'ngpts':
            val = self.ulm.density.density.shape
        else:
            val = self.ulm.asdict().get(key, None)
        return val

    def get_program_version(self):
        return self.ulm.gpaw_version

    def get_smearing_width(self):
        if self.get_parameter('occupations') is None:
            return 0.0 if tuple(self.get_parameter('kpts')) == (1, 1, 1) else 0.1
        else:
            return self.get_parameter('occupations').get('width')


class GPAWParser(FairdiParser):
    def __init__(self):
        super(). __init__(
            name='parsers/gpaw', code_name='GPAW', code_homepage='https://wiki.fysik.dtu.dk/gpaw/',
            mainfile_name_re=(r'^.*\.(gpw2|gpw)$'),
            mainfile_mime_re=r'application/(x-tar|octet-stream)')

        self._metainfo_env = m_env
        self.gpw_parser = GPWParser()
        self.gpw2_parser = GPW2Parser()
        self._xc_map = {
            'LDA': ['LDA_X', 'LDA_C_PW'],
            'PW91': ['GGA_X_PW91', 'GGA_C_PW91'],
            'PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'PBEsol': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
            'revPBE': ['GGA_X_PBE_R', 'GGA_C_PBE'],
            'RPBE': ['GGA_X_RPBE', 'GGA_C_PBE'],
            'BLYP': ['GGA_X_B88', 'GGA_C_LYP'],
            'HCTH407': ['GGA_XC_HCTH_407'],
            'WC': ['GGA_X_WC', 'GGA_C_PBE'],
            'AM05': ['GGA_X_AM05', 'GGA_C_AM05'],
            'M06-L': ['MGGA_X_M06_L', 'MGGA_C_M06_L'],
            'TPSS': ['MGGA_X_TPSS', 'MGGA_C_TPSS'],
            'revTPSS': ['MGGA_X_REVTPSS', 'MGGA_C_REVTPSS'],
            'mBEEF': ['MGGA_X_MBEEF', 'GGA_C_PBE_SOL']}

    def init_parser(self, filepath, logger):
        self.parser = self.gpw_parser
        self.parser.mainfile = filepath
        if self.parser.mainfile_obj is None:
            self.parser = self.gpw2_parser
            self.parser.mainfile = filepath
        self.parser.logger = logger

    def apply_unit(self, val, unit):
        units_map = {
            'ev': ureg.eV, 'hartree': ureg.hartree, 'angstrom': ureg.angstrom,
            'bohr': ureg.bohr, 'femtosecond': ureg.fs}
        p_unit = self.parser.info['parameter'].get(unit, '').lower()
        unit = units_map.get(p_unit, p_unit) if p_unit else unit
        return val * unit

    def get_basis_set_name(self):
        basis_set = self.archive.section_run[-1].program_basis_set_type
        if basis_set == 'plane waves':
            pw_cutoff = self.parser.get_parameter('planewavecutoff')
            pw_cutoff = self.apply_unit(pw_cutoff, 'energyunit')
            return 'PW_%.1f_Ry' % (pw_cutoff.to('rydberg').magnitude)
        elif basis_set == 'real space grid':
            cell = self.parser.get_array('unitcell')
            ngpts = self.parser.get_array_dimension('ngpts')
            if cell is None or ngpts is None:
                return
            h_grid = np.linalg.norm(cell, axis=1) / np.array(ngpts[:3])
            h_grid = self.apply_unit(np.sum(h_grid) / 3.0, 'lengthunit')
            return 'GR_%.1f' % (h_grid.to('fm').magnitude)
        elif basis_set == 'numeric AOs':
            return self.parser.get_parameter('basisset')

    def get_mode(self):
        mode = self.parser.get_parameter('mode')
        if isinstance(mode, dict):
            mode = mode.get('name', None)
        return mode

    def get_nspin(self):
        # TODO another way determine spin?
        magnetic_moments = self.parser.get_array('magneticmoments')
        return 1 if magnetic_moments is None else 2

    def get_fermi_level(self):
        fermi_level = [self.parser.get_parameter('fermilevel')] * self.get_nspin()
        if None in fermi_level:
            return
        split = self.parser.get_parameter('split')
        if split is not None:
            fermi_level = [v + (-i + 0.5) * split for i, v in enumerate(fermi_level)]
        return self.apply_unit(fermi_level, 'energyunit')

    def parse_method(self):
        sec_method = self.archive.section_run[-1].m_create(Method)
        sec_method.relativity_method = 'pseudo_scalar_relativistic'
        sec_method.electronic_structure_method = 'DFT'

        threshold_energy = self.parser.get_parameter('energyerror')
        sec_method.scf_threshold_energy_change = self.apply_unit(threshold_energy, 'energyunit')

        smearing_width = self.parser.get_smearing_width()
        if smearing_width is not None:
            sec_method.smearing_kind = 'fermi'
            sec_method.smearing_width = self.apply_unit(
                smearing_width, 'energyunit').to('joule').magnitude

        charge = self.parser.get_parameter('charge')
        if charge is not None:
            sec_method.total_charge = int(charge)

        xc_functional = self.parser.get_parameter('xcfunctional')
        for xc in self._xc_map.get(xc_functional, [xc_functional]):
            sec_xc_functionals = sec_method.m_create(XCFunctionals)
            sec_xc_functionals.XC_functional_name = xc

        method_keys = [
            'fix_magnetic_moment', 'fix_density', 'density_convergence_criterion',
            'mix_class', 'mix_beta', 'mix_weight', 'mix_old', 'maximum_angular_momentum',
            'symmetry_time_reversal_switch']
        for key in method_keys:
            val = self.parser.get_parameter(key.replace('_', ''))
            if val is None:
                continue
            setattr(sec_method, 'x_gpaw_%s' % key, val)

    def parse_scc(self):
        sec_run = self.archive.section_run[-1]
        sec_scc = sec_run.m_create(SingleConfigurationCalculation)

        # energies (in gpw, energies are part of parameters)
        energy_keys = [
            'energy_total', 'energy_free', 'energy_XC', 'electronic_kinetic_energy',
            'energy_correction_entropy']
        for key in energy_keys:
            val = self.parser.get_parameter(key)
            if val is not None:
                setattr(sec_scc, key, self.apply_unit(val, 'energyunit'))

        # forces
        forces_key = ['atom_forces_free', 'atom_forces_free_raw']
        for key in forces_key:
            val = self.parser.get_array(key)
            if val is not None:
                energyunit = self.apply_unit(1, 'energyunit').units
                lengthunit = self.apply_unit(1, 'lengthunit').units
                setattr(sec_scc, key, val * energyunit / lengthunit)

        # magnetic moments
        magnetic_moments = self.parser.get_array('magneticmoments')
        if magnetic_moments is not None:
            sec_scc.x_gpaw_magnetic_moments = magnetic_moments
            sec_scc.x_gpaw_fixed_spin_Sz = magnetic_moments.sum() / 2.

        # fermi level
        fermi_level = self.get_fermi_level()
        if fermi_level is not None:
            sec_scc.energy_reference_fermi = fermi_level

        # eigenvalues
        eigenvalues = self.parser.get_array('eigenvalues')
        if eigenvalues is not None:
            sec_eigenvalues = sec_scc.m_create(Eigenvalues)
            sec_eigenvalues.eigenvalues_kind = 'normal'
            sec_eigenvalues.eigenvalues_values = self.apply_unit(eigenvalues, 'energyunit')
            for key in ['occupation', 'kpoints']:
                val = self.parser.get_array(key)
                if val is None:
                    continue
                setattr(sec_eigenvalues, 'eigenvalues_%s' % key, val)

        # band path (TODO only in ulm?)
        band_paths = self.parser.get_array('band_paths')
        if band_paths is not None:
            sec_k_band = sec_scc.m_create(KBand)
            for band_path in band_paths:
                sec_band_seg = sec_k_band.m_create(KBandSegment)
                if band_path.get('eigenvalues', None) is not None:
                    sec_band_seg.band_energies = self.apply_unit(band_path.get(
                        'eigenvalues'), 'energyunit')
                kpoints = band_path.get('kpoints', None)
                if kpoints is not None:
                    sec_band_seg.band_k_points = kpoints
                    sec_band_seg.band_segm_start_end = np.asarray(kpoints[0], kpoints[-1])
                if band_path.get('labels', None) is not None:
                    sec_band_seg.band_segm_labels = band_path.get('labels')

        # volumetric data
        density = self.parser.get_array('density')
        if density is not None:
            cell = self.parser.get_array('unitcell')
            origin = -0.5 * cell.sum(axis=0)
            pbc = self.parser.get_array('boundaryconditions')
            npoints = np.array(density.shape[1:])
            npoints = [(npt + 1) if not pbc[i] else npt for i, npt in enumerate(npoints)]
            displacements = cell / np.array(npoints)
            lengthunit = self.apply_unit(1, 'lengthunit').units
            energyunit = self.apply_unit(1, 'energyunit').units
            for key in ['density', 'potential_effective']:
                val = self.parser.get_array(key)
                if val is None:
                    continue
                sec_vol = sec_scc.m_create(VolumetricData)
                sec_vol.volumetric_data_kind = key
                sec_vol.volumetric_data_origin = (origin * lengthunit).to('m').magnitude
                sec_vol.volumetric_data_displacements = (displacements * lengthunit).to('m').magnitude
                if key == 'density':
                    val = (val / lengthunit ** 3).to('1/m**3')
                else:
                    val = (val * energyunit / lengthunit ** 3).to('J/m**3')
                sec_vol.volumetric_data_values = val.magnitude

        converged = self.parser.get_parameter('converged')
        if converged is not None:
            sec_scc.single_configuration_calculation_converged = converged

        sec_scc.single_configuration_calculation_to_system_ref = sec_run.section_system[-1]
        sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]

    def parse_system(self):
        sec_system = self.archive.section_run[-1].m_create(System)

        cell = self.parser.get_array('unitcell')
        if cell is not None:
            cell = self.apply_unit(cell, 'lengthunit')
            sec_system.lattice_vectors = cell
            sec_system.simulation_cell = cell

        sec_system.atom_labels = [
            ase.data.chemical_symbols[z] for z in self.parser.get_array('atomicnumbers')]

        positions = self.parser.get_array('atom_positions')
        sec_system.atom_positions = self.apply_unit(positions, 'lengthunit')

        pbc = [True, True, True] if self.get_mode() == 'pw' else np.array(
            self.parser.get_array('boundaryconditions'), bool)
        sec_system.configuration_periodic_dimensions = pbc

        momenta = self.parser.get_array('momenta')
        if momenta is not None:
            masses = np.array([ase.data.atomic_masses[self.parser.get_array('atomicnumbers')]])
            velocities = momenta / masses.reshape(-1, 1)
            sec_system.atom_velocities = velocities * ase.units.fs / ase.units.Angstrom * ureg.angstrom / ureg.fs

    def parse(self, filepath, archive, logger):
        self.filepath = filepath
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.init_parser(filepath, logger)

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'GPAW'
        sec_run.program_version = self.parser.get_program_version()

        mode = self.get_mode()
        if mode == 'pw':
            sec_run.program_basis_set_type = 'plane waves'
            sec_basis = sec_run.m_create(BasisSetCellDependent)
            pw_cutoff = self.parser.get_parameter('planewavecutoff')
            pw_cutoff = self.apply_unit(pw_cutoff, 'energyunit')
            sec_basis.basis_set_planewave_cutoff = pw_cutoff
            sec_basis.basis_set_cell_dependent_name = self.get_basis_set_name()
        elif mode == 'fd':
            sec_basis = sec_run.m_create(BasisSetCellDependent)
            sec_run.program_basis_set_type = 'real space grid'
            sec_basis.basis_set_cell_dependent_name = self.get_basis_set_name()
        elif mode == 'lcao':
            sec_run.program_basis_set_type = 'numeric AOs'
            sec_basis = sec_run.m_create(BasisSetAtomCentered)
            sec_basis.basis_set_atom_centered_short_name = self.get_basis_set_name()

        sec_sampling_method = sec_run.m_create(SamplingMethod)
        sec_sampling_method.sampling_method = 'geometry_optimization'
        sec_sampling_method.ensemble_type = 'NVE'

        self.parse_method()

        self.parse_system()

        self.parse_scc()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup, find_packages
from subprocess import call
from setuptools.command.install import install as setup_install
import runpy
import os
import sys
import json
import re
import fastentrypoints  # pylint: disable=unused-import

'''
This setup.py works differently for creating a distribution than installing for
development. The idea is that for distributions, we compile the requirements, packages, and
data files from all dependencies and install everything under one package called 'nomad-lab'.
For development we install each dependency as its own project. This is mainly due to
pip's limitation of not being able to install from multiple source folders in develop
mode.

The compilation process can by run with ``python setup.py compile``. It will create a
file with setup kwargs calles ``./setup.json``. If this file exists setup will be called
with compiled args. If this file does not exist, setup will be called with regular args.

This setup makes use of ``extras_require``. The extras are parsing, infrastructure, dev,
and all. Where all is the union of the other ones. The extras are parsed from the
requirements.txt where specific comments are used to assign an extra to requirements.
'''


ignore_extra_requires = ['optimade']
''' Dependencies where the extra_requires should not be added '''


def parse_requirements():
    '''
    Parses the requirements.txt file to extras install and extra requirements.
    Sections headed with # [extra] are assigned to the 'extra' extra.

    Returns:
        Tuple with install and extra requires passible to :func:`setuptools.setup`.
    '''
    with open('requirements.txt', 'rt') as f:
        lines = f.readlines()

    extras_require = {}
    requires = []
    all_requires = []
    current = None
    for line in lines:
        line = line.strip()

        if line == '':
            continue

        match = re.match(r'^#\s*\[([a-zA-Z0-9_]+)\]$', line)
        if match:
            extra = match.group(1)
            current = list()
            extras_require[extra] = current
        elif line.startswith('#'):
            continue
        else:
            line = line.split('#')[0].strip()
            if current is None:
                requires.append(line)
            else:
                current.append(line)
                all_requires.append(line)

    extras_require['all'] = all_requires

    return requires, extras_require


class install(setup_install):
    def __post_install(self, dir):
        if os.name == 'posix':
            call(['./auto_complete_install.sh'])

    def run(self):
        setup_install.run(self)
        self.execute(self.__post_install, (self.install_lib, ), msg='installing autocompletion')


def compile_dependency_setup_kwargs(paths, **kwargs):
    import setuptools
    import distutils.core as distutils_core

    # collect all kwargs from all setup.pys
    results = {}
    current = {}

    def patched_setup(*args, **kwargs):
        assert len(args) == 0
        assert current['name'] not in results, 'current is %s' % current['name']

        results[current['name']] = {
            'meta': dict(**current),
            'kwargs': kwargs
        }

    if len(kwargs) > 0:
        current['name'] = kwargs.get('name', 'nomad')
        current['directory'] = './'
        current['setup.py'] = './setup.py'
        patched_setup(**kwargs)

    setuptoolss = [setuptools, distutils_core]
    orig_setups = []
    for st in setuptoolss:
        orig_setups.append((st, getattr(st, 'setup')))
        setattr(st, 'setup', patched_setup)

    for path in paths:
        for root, _, files in os.walk(path):
            for file in files:
                if os.path.basename(file) == 'setup.py':
                    setup_path = os.path.join(root, file)
                    current['name'] = os.path.basename(os.path.dirname(setup_path))
                    current['directory'] = os.path.dirname(setup_path)
                    current['setup.py'] = setup_path
                    cwd = os.getcwd()
                    os.chdir(os.path.dirname(setup_path))
                    try:
                        runpy.run_path(file, run_name='__main__')
                    except Exception:
                        import traceback
                        traceback.print_exc()
                        print('Could not run %s' % setup_path)
                        sys.exit(1)
                    finally:
                        os.chdir(cwd)

    for st, setup in orig_setups:
        setattr(st, 'setup', setup)

    # combine the kwargs
    all_packages = []
    all_package_dir = {}
    all_package_data = {}
    all_install_requires = {}
    all_names = set()
    for _, setup_data in results.items():
        meta = setup_data['meta']
        local_kwargs = setup_data['kwargs']

        if 'name' in local_kwargs:
            all_names.add(local_kwargs['name'])

        name = local_kwargs.get('name', meta['name'])

        # 1. packages, package_dir
        package_dir = local_kwargs.get('package_dir', {'': './'})
        packages = local_kwargs.get('packages', [])

        assert len(package_dir) == 1
        assert '' in package_dir

        for package in packages:
            root = package.split('.')[0]
            if root not in package_dir and root not in all_package_dir:
                all_package_dir[root] = os.path.normpath(
                    os.path.join(meta['directory'], package_dir[''], root))
            all_packages.append(package)

        # 2. package_data
        all_package_data.update(**local_kwargs.get('package_data', {}))

        # 3. requires
        local_install_requires = set()
        if name not in ignore_extra_requires:
            for extra_require in local_kwargs.get('extras_require', {}).values():
                for require in extra_require:
                    local_install_requires.add(require)

        for require in local_kwargs.get('install_requires', []):
            local_install_requires.add(require)
        all_install_requires[name] = local_install_requires

    # automatically add parser deps
    for _, setup_data in results.items():
        if 'parsers' in setup_data['meta']['setup.py']:
            parsing = kwargs.setdefault('extras_require', {}).setdefault('parsing', [])
            all = kwargs.setdefault('extras_require', {}).setdefault('all', [])
            for require in setup_data['kwargs'].get('install_requires', []):
                if require not in parsing:
                    parsing.append(require)
                    all.append(require)

    def replace_own_packages(requires):
        ''' replaces nomad dependencies with their requirements '''
        for other in all_names:
            if other in requires:
                requires.remove(other)
                requires.extend(all_install_requires[other])

        # remove dups
        sorted_requires = sorted(requires)
        sorted_normalized = [re.match(r'^([a-zA-Z0-9_\.\-]*).*$', r).group(1) for r in sorted_requires]
        to_remove = []
        for i, require in enumerate(sorted_normalized):
            if i + 1 < len(sorted_requires) and require == sorted_normalized[i + 1]:
                to_remove.append(i)
        for i in to_remove:
            requires.remove(sorted_requires[i])

    # run twice because dependencies can be dependencies of dependencies
    for _ in range(0, 2):
        for extra_require in kwargs.get('extras_require', {}).values():
            replace_own_packages(extra_require)
        replace_own_packages(kwargs.get('install_requires', []))

    kwargs.update(**{
        'package_dir': all_package_dir,
        'packages': all_packages,
        'package_data': all_package_data
    })

    return kwargs


def setup_kwargs():
    from nomad import config

    install_requires, extras_require = parse_requirements()
    with open("README.md", "r") as fh:
        long_description = fh.read()

    return dict(
        name='nomad-lab',
        author='NOMAD Laboratory',
        author_email='markus.scheidgen@physik.hu-berlin.de',
        url='https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-FAIR',
        version=config.meta.version,
        license='APACHE 2.0',
        description='The NOvel MAterials Discovery (NOMAD) Python package',
        long_description=long_description,
        long_description_content_type="text/markdown",
        package_dir={'': './'},
        packages=['nomad.%s' % pkg for pkg in find_packages('./nomad')] + ['nomad'],
        setup_requires=['pip', 'setuptools', 'wheel', 'fastentrypoints', 'numpy', 'pyyaml'],
        install_requires=install_requires,
        extras_require=extras_require,
        include_package_data=True,
        python_requires='>=3.6',
        entry_points='''
            [console_scripts]
            nomad=nomad.cli:run_cli
        ''')


if __name__ == '__main__':
    if len(sys.argv) == 2 and sys.argv[1] == 'compile':
        kwargs = compile_dependency_setup_kwargs(['dependencies'], **setup_kwargs())
        kwargs['package_data']['optimade.grammar'] = ['*.lark']
        with open('setup.json', 'wt') as f:
            json.dump(kwargs, f, indent=2)
        sys.exit(0)

    if len(sys.argv) == 2 and sys.argv[1] == 'dry':
        import pprint
        pprint.pprint(
            compile_dependency_setup_kwargs(['dependencies'], **setup_kwargs()))
        sys.exit(0)

    if os.path.exists('setup.json'):
        with open('setup.json', 'rt') as f:
            kwargs = json.load(f)

    else:
        kwargs = setup_kwargs()

    setup(cmdclass={'install': install}, **kwargs)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

from nomad.metainfo import (
    Section, Quantity, MSection, SubSection, SectionProxy, Package
)
from nomad.datamodel.metainfo import public


m_package = Package(
    name='lobster_nomadmetainfo_json',
    description='None')


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_lobster_abs_total_spilling = Quantity(
        type=float,
        shape=['number_of_spin_channels'],
        description='''
        Absolute total spilling (in all levels)
        when projecting from the original wave functions into the local basis.
        ''')

    x_lobster_abs_charge_spilling = Quantity(
        type=float,
        shape=['number_of_spin_channels'],
        description='''
        Absolute total spilling of density (in occupied levels)
        when projecting from the original wave functions into the local basis.
        ''')

    x_lobster_section_cohp = SubSection(
        sub_section=SectionProxy('x_lobster_section_cohp'))

    x_lobster_section_coop = SubSection(
        sub_section=SectionProxy('x_lobster_section_coop'))

    x_lobster_section_atom_projected_dos = SubSection(
        sub_section=SectionProxy('x_lobster_section_atom_projected_dos'),
        repeats=True)


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True)

    x_lobster_code = Quantity(
        type=str,
        description='''
        Used PAW program
        ''')


class x_lobster_section_cohp(MSection):
    """
    This is a section containing the crystal orbital hamilton population (COHP)
    and integrated COHP (iCOHP) values.
    """
    m_def = Section(validate=False)

    x_lobster_number_of_cohp_pairs = Quantity(
        type=int,
        description='''
        Number of atom pairs for which are the COHPs and iCOHPs calculated.
        ''')

    x_lobster_cohp_atom1_labels = Quantity(
        type=str,
        shape=['x_lobster_number_of_cohp_pairs'],
        description='''
        Species and indices of the first atom for which is the specific COHP/iCOHP calculated
        ''')

    x_lobster_cohp_atom2_labels = Quantity(
        type=str,
        shape=['x_lobster_number_of_cohp_pairs'],
        description='''
        Species and indices of the second atom for which is the specific COHP/iCOHP calculated
        ''')

    x_lobster_cohp_distances = Quantity(
        type=np.dtype(np.float64),
        unit='meter',
        shape=['x_lobster_number_of_cohp_pairs'],
        description='''
        Distance between atoms of the pair for which is the specific COHP/iCOHP calculated.
        ''')

    x_lobster_cohp_translations = Quantity(
        type=np.dtype(np.int8),
        shape=['x_lobster_number_of_cohp_pairs', 3],
        description='''
        Vector connecting the unit-cell of the first atom with the one of the second atom

        This is only used with LOBSTER versions 3.0.0 and above, older versions use
        x_lobster_cohp_number_of_bonds instead.
        ''')

    x_lobster_integrated_cohp_at_fermi_level = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['number_of_spin_channels', 'x_lobster_number_of_cohp_pairs'],
        description='''
        Calculated iCOHP values ingregrated up to the Fermi level.
        ''')

    x_lobster_number_of_cohp_values = Quantity(
        type=int,
        description='''
        Number of energy values for the COHP and iCOHP.
        ''')

    x_lobster_cohp_energies = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['x_lobster_number_of_cohp_values'],
        description='''
        Array containing the set of discrete energy values for COHP and iCOHP.
        ''')

    x_lobster_cohp_values = Quantity(
        type=np.dtype(np.float32),
        shape=['x_lobster_number_of_cohp_pairs', 'number_of_spin_channels', 'x_lobster_number_of_cohp_values'],
        description='''
        Calculated COHP values.
        ''')

    x_lobster_integrated_cohp_values = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['x_lobster_number_of_cohp_pairs', 'number_of_spin_channels', 'x_lobster_number_of_cohp_values'],
        description='''
        Calculated iCOHP values.
        ''')

    x_lobster_average_cohp_values = Quantity(
        type=np.dtype(np.float32),
        shape=['number_of_spin_channels', 'x_lobster_number_of_cohp_values'],
        description='''
        Calculated COHP values averaged over all pairs.
        ''')

    x_lobster_average_integrated_cohp_values = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['number_of_spin_channels', 'x_lobster_number_of_cohp_values'],
        description='''
        Calculated iCOHP values averaged over all pairs.
        ''')

    x_lobster_cohp_number_of_bonds = Quantity(
        type=int,
        shape=['x_lobster_number_of_cohp_pairs'],
        description='''
        Number of bonds between first atom and the second atom (including
        the periodic images).

        This is only used in older LOBSTER versions, new versions print one line
        for every neighbor, so a pair which had x_lobster_icohp_number_of_bonds = 4
        in the old version would actually show as 4 lines in the ICOHPLIST or 4 columns
        in the COPHCAR in the new format.
        ''')


class x_lobster_section_coop(MSection):
    """
    This is a section containing the crystal orbital hamilton population (COOP)
    and integrated coop (iCOOP) values.
    """
    m_def = Section(validate=False)

    x_lobster_number_of_coop_pairs = Quantity(
        type=int,
        description='''
        Number of atom pairs for which are the COOPs and iCOOPs calculated.
        ''')

    x_lobster_coop_atom1_labels = Quantity(
        type=str,
        shape=['x_lobster_number_of_coop_pairs'],
        description='''
        Species and indices of the first atom for which is the specific COOP/iCOOP calculated
        ''')

    x_lobster_coop_atom2_labels = Quantity(
        type=str,
        shape=['x_lobster_number_of_coop_pairs'],
        description='''
        Species and indices of the second atom for which is the specific COOP/iCOOP calculated
        ''')

    x_lobster_coop_distances = Quantity(
        type=np.dtype(np.float64),
        unit='meter',
        shape=['x_lobster_number_of_coop_pairs'],
        description='''
        Distance between atoms of the pair for which is the specific COOP/iCOOP calculated.
        ''')

    x_lobster_coop_translations = Quantity(
        type=np.dtype(np.int8),
        shape=['x_lobster_number_of_coop_pairs', 3],
        description='''
        Vector connecting the unit-cell of the first atom with the one of the second atom

        This is only used with LOBSTER versions 3.0.0 and above, older versions use
        x_lobster_coop_number_of_bonds instead.
        ''')

    x_lobster_integrated_coop_at_fermi_level = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['number_of_spin_channels', 'x_lobster_number_of_coop_pairs'],
        description='''
        Calculated iCOOP values ingregrated up to the Fermi level.
        ''')

    x_lobster_number_of_coop_values = Quantity(
        type=int,
        description='''
        Number of energy values for the COOP and iCOOP.
        ''')

    x_lobster_coop_energies = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['x_lobster_number_of_coop_values'],
        description='''
        Array containing the set of discrete energy values for COOP and iCOOP.
        ''')

    x_lobster_coop_values = Quantity(
        type=np.dtype(np.float32),
        shape=['x_lobster_number_of_coop_pairs', 'number_of_spin_channels', 'x_lobster_number_of_coop_values'],
        description='''
        Calculated COOP values.
        ''')

    x_lobster_integrated_coop_values = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['x_lobster_number_of_coop_pairs', 'number_of_spin_channels', 'x_lobster_number_of_coop_values'],
        description='''
        Calculated iCOOP values.
        ''')

    x_lobster_average_coop_values = Quantity(
        type=np.dtype(np.float32),
        shape=['number_of_spin_channels', 'x_lobster_number_of_coop_values'],
        description='''
        Calculated COOP values averaged over all pairs.
        ''')

    x_lobster_average_integrated_coop_values = Quantity(
        type=np.dtype(np.float32),
        unit='joule',
        shape=['number_of_spin_channels', 'x_lobster_number_of_coop_values'],
        description='''
        Calculated iCOOP values averaged over all pairs.
        ''')

    x_lobster_coop_number_of_bonds = Quantity(
        type=int,
        shape=['x_lobster_number_of_coop_pairs'],
        description='''
        Number of bonds between first atom and the second atom (including
        the periodic images).

        This is only used in older LOBSTER versions, new versions print one line
        for every neighbor, so a pair which had x_lobster_icoop_number_of_bonds = 4
        in the old version would actually show as 4 lines in the ICOOPLIST or 4 columns
        in the COOPCAR in the new format.
        ''')


class x_lobster_section_atom_projected_dos(MSection):
    '''
    Section collecting the information on an atom projected density of states (DOS)
    evaluation.
    FIXME: this should ultimatelly go into some common section but that is not possible
    right now, see: https://matsci.org/t/section-atom-projected-dos/36008
    '''

    m_def = Section(validate=False)

    x_lobster_atom_projected_dos_energies = Quantity(
        type=np.dtype(np.float64),
        shape=['x_lobster_number_of_atom_projected_dos_values'],
        unit='joule',
        description='''
        Array containing the set of discrete energy values for the atom-projected density
        (electronic-energy) of states (DOS).
        ''')

    x_lobster_atom_projected_dos_lm = Quantity(
        type=np.dtype(np.int32),
        shape=['x_lobster_number_of_lm_atom_projected_dos', 2],
        description='''
        Tuples of $l$ and $m$ values for which x_lobster_atom_projected_dos_values_lm are given.
        For the quantum number $l$ the conventional meaning of azimuthal quantum number is
        always adopted. For the integer number $m$, besides the conventional use as
        magnetic quantum number ($l+1$ integer values from $-l$ to $l$), a set of
        different conventions is accepted (see the [m_kind wiki
        page](https://gitlab.rzg.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/m-kind).
        The adopted convention is specified by atom_projected_dos_m_kind.
        ''')

    x_lobster_atom_projected_dos_m_kind = Quantity(
        type=str,
        description='''
        String describing what the integer numbers of $m$ in atom_projected_dos_lm mean.
        The allowed values are listed in the [m_kind wiki
        page](https://gitlab.rzg.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/m-kind).
        ''')

    x_lobster_atom_projected_dos_values_lm = Quantity(
        type=np.dtype(np.float64),
        shape=['x_lobster_number_of_lm_atom_projected_dos', 'number_of_spin_channels',
               'x_lobster_number_of_atom_projected_dos_values'],
        description='''
        Values correspond to the number of states for a given energy (the set of discrete
        energy values is given in atom_projected_dos_energies) divided into contributions
        from each $l,m$ channel for the atom-projected density (electronic-energy) of
        states for atom specified in atom_projected_dos_atom_index.
        ''')

    x_lobster_atom_projected_dos_values_total = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'x_lobster_number_of_atom_projected_dos_values'],
        description='''
        Values correspond to the number of states for a given energy (the set of discrete
        energy values is given in atom_projected_dos_energies) divided into contributions
        summed up over all $l$ channels for the atom-projected density (electronic-energy)
        of states (DOS) for atom specified in atom_projected_dos_atom_index.
        ''')

    x_lobster_number_of_atom_projected_dos_values = Quantity(
        type=int,
        description='''
        Gives the number of energy values for the atom-projected density of states (DOS)
        based on x_lobster_atom_projected_dos_values_lm and
        x_lobster_atom_projected_dos_values_total.
        ''')

    x_lobster_number_of_lm_atom_projected_dos = Quantity(
        type=int,
        description='''
        Gives the number of $l$, $m$ combinations for the atom projected density of states
        (DOS) defined in x_lobster_section_atom_projected_dos.
        ''')

    x_lobster_atom_projected_dos_atom_index = Quantity(
        type=int,
        description='''
        Index of atom for which is the x_lobster_atom_projected_dos provided.
        ''')


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import phonopy
from phonopy.units import THzToEv
from phonopy.structure.atoms import PhonopyAtoms

from phonopyparser.phonopy_properties import PhononProperties

import nomad.config
from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, System, SystemToSystemRefs, Method,\
    SingleConfigurationCalculation, KBand, KBandSegment, Dos, FrameSequence,\
    ThermodynamicalProperties, SamplingMethod, Workflow, Phonon, CalculationToCalculationRefs

from nomad.parsing.parser import FairdiParser
from .metainfo import m_env


def read_aims(filename):
    '''Method to read FHI-aims geometry files in phonopy context.'''
    cell = []
    positions = []
    fractional = []
    symbols = []
    magmoms = []
    with open(filename) as f:
        while True:
            line = f.readline()
            if not line:
                break
            line = line.split()
            if line[0] == 'lattice_vector':
                cell.append([float(x) for x in line[1:4]])
            elif line[0].startswith('atom'):
                fractional.append(line[0] == 'atom_frac')
                positions.append([float(x) for x in line[1:4]])
                symbols.append(line[4])
            elif line[0] == 'initial_moment':
                magmoms.append(float(line[1]))

    for n, pos in enumerate(positions):
        if fractional[n]:
            positions[n] = [sum([pos[j] * cell[j][i] for j in range(3)]) for i in range(3)]
    if len(magmoms) == len(positions):
        return PhonopyAtoms(cell=cell, symbols=symbols, positions=positions, magmoms=magmoms)
    else:
        return PhonopyAtoms(cell=cell, symbols=symbols, positions=positions)


class Atoms_with_forces(PhonopyAtoms):
    ''' Hack to phonopy.atoms to maintain ASE compatibility also for forces.'''

    def get_forces(self):
        return self.forces


def read_aims_output(filename):
    ''' Read FHI-aims output
        returns geometry with forces from last self-consistency iteration'''
    cell = []
    symbols = []
    positions = []
    forces = []
    N = 0

    with open(filename) as f:
        while True:
            line = f.readline()
            if not line:
                break
            if 'Number of atoms' in line:
                N = int(line.split()[5])
            elif '| Unit cell:' in line:
                cell = [[float(x) for x in f.readline().split()[1:4]] for _ in range(3)]
            elif 'Atomic structure:' in line or 'Updated atomic structure:' in line:
                positions = []
                symbols = []
                symbol_index = 3 if 'Atomic' in line else 4
                position_index = 4 if 'Atomic' in line else 1
                while len(positions) != N:
                    line = f.readline()
                    if 'Species' in line or 'atom ' in line:
                        line = line.split()
                        positions.append([float(x) for x in line[position_index:position_index + 3]])
                        symbols.append(line[symbol_index])
            elif 'Total atomic forces' in line:
                forces = [[float(x) for x in f.readline().split()[2:5]] for _ in range(N)]

    atoms = Atoms_with_forces(cell=cell, symbols=symbols, positions=positions)
    atoms.forces = forces

    return atoms


def read_forces_aims(reference_supercells, tolerance=1E-6, logger=None):
    '''
    Collect the pre calculated forces for each of the supercells
    '''
    def get_aims_output_file(directory):
        files = [f for f in os.listdir(directory) if f.endswith('.out')]
        output = None
        for f in files:
            try:
                output = read_aims_output(os.path.join(directory, f))
                break
            except Exception:
                pass
        return output

    def is_equal(reference, calculated):
        if len(reference) != len(calculated):
            logger.warn('Inconsistent number of atoms.')
            return False
        if (reference.get_atomic_numbers() != calculated.get_atomic_numbers()).any():
            logger.warn('Inconsistent species.')
            return False
        if (abs(reference.get_cell() - calculated.get_cell()) > tolerance).any():
            logger.warn('Inconsistent cell.')
            return False
        ref_pos = reference.get_scaled_positions()
        cal_pos = calculated.get_scaled_positions()
        # wrap to bounding cell
        ref_pos %= 1.0
        cal_pos %= 1.0
        if (abs(ref_pos - cal_pos) > tolerance).any():
            logger.warn('Inconsistent positions.')
            return False
        return True

    reference_paths, forces_sets = [], []

    n_pad = int(np.ceil(np.log10(len(reference_supercells) + 1))) + 1
    for n, reference_supercell in enumerate(reference_supercells):
        directory = 'phonopy-FHI-aims-displacement-%s' % (str(n + 1).zfill(n_pad))
        filename = os.path.join(directory, '%s.out' % directory)
        if os.path.isfile(filename):
            calculated_supercell = read_aims_output(filename)
        else:
            # try reading out files
            calculated_supercell = get_aims_output_file(directory)

        # compare if calculated cell really corresponds to supercell
        if not is_equal(reference_supercell, calculated_supercell):
            logger.error('Supercells do  not match')

        forces = np.array(calculated_supercell.get_forces())
        drift_force = forces.sum(axis=0)
        for force in forces:
            force -= drift_force / forces.shape[0]
        forces_sets.append(forces)
        reference_paths.append(filename)
    return forces_sets, reference_paths


class ControlParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_nac(val_in):
            val = val_in.strip().split()
            nac = dict(file=val[0], method=val[1].lower())
            if len(val) > 2:
                nac['delta'] = [float(v) for v in val[3:6]]
            return nac

        def str_to_supercell(val_in):
            val = [int(v) for v in val_in.strip().split()]
            if len(val) == 3:
                return np.diag(val)
            else:
                return np.reshape(val, (3, 3))

        self._quantities = [
            Quantity('displacement', r'\n *phonon displacement\s*([\d\.]+)', dtype=float),
            Quantity('symmetry_thresh', r'\n *phonon symmetry_thresh\s*([\d\.]+)', dtype=float),
            Quantity('frequency_unit', r'\n *phonon frequency_unit\s*(\S+)'),
            Quantity('supercell', r'\n *phonon supercell\s*(.+)', str_operation=str_to_supercell),
            Quantity('nac', r'\n *phonon nac\s*(.+)', str_operation=str_to_nac)]


class PhonopyParser(FairdiParser):
    def __init__(self, **kwargs):
        super().__init__(
            name='parsers/phonopy', code_name='Phonopy', code_homepage='https://phonopy.github.io/phonopy/',
            mainfile_name_re=(r'(.*/phonopy-FHI-aims-displacement-0*1/control.in$)|(.*/phon.+yaml)')
        )
        self._kwargs = kwargs
        self.control_parser = ControlParser()

    @property
    def mainfile(self):
        return self._filepath

    @mainfile.setter
    def mainfile(self, val):
        self._phonopy_obj = None
        self.references = []
        self._filepath = os.path.abspath(val)

    @property
    def calculator(self):
        if 'control.in' in self.mainfile:
            return 'fhi-aims'
        elif self.mainfile.endswith('.yaml'):
            return 'vasp'

    @property
    def phonopy_obj(self):
        if self._phonopy_obj is None:
            if self.calculator == 'fhi-aims':
                self._build_phonopy_object_fhi_aims()
            elif self.calculator == 'vasp':
                self._build_phonopy_object_vasp()
        return self._phonopy_obj

    def _build_phonopy_object_vasp(self):
        cwd = os.getcwd()
        os.chdir(os.path.dirname(self.mainfile))

        try:
            phonopy_obj = phonopy.load(self.mainfile)
        finally:
            os.chdir(cwd)

        self._phonopy_obj = phonopy_obj

    def _build_phonopy_object_fhi_aims(self):
        cwd = os.getcwd()
        os.chdir(os.path.dirname(os.path.dirname(self.mainfile)))

        try:
            cell_obj = read_aims('geometry.in')
            self.control_parser.mainfile = 'control.in'
            supercell_matrix = self.control_parser.get('supercell')
            displacement = self.control_parser.get('displacement', 0.001)
            sym = self.control_parser.get('symmetry_thresh', 1e-6)
            try:
                phonopy_obj = phonopy.Phonopy(cell_obj, supercell_matrix, symprec=sym)
                phonopy_obj.generate_displacements(distance=displacement)
                supercells = phonopy_obj.get_supercells_with_displacements()
                set_of_forces, relative_paths = read_forces_aims(supercells, logger=self.logger)
            except Exception:
                self.logger.error("Error generating phonopy object.")
                set_of_forces = []
                phonopy_obj = None
                relative_paths = []

            prep_path = self.mainfile.split("phonopy-FHI-aims-displacement-")
            # Try to resolve references as paths relative to the upload root.
            try:
                for path in relative_paths:
                    abs_path = "%s%s" % (prep_path[0], path)
                    rel_path = abs_path.split(nomad.config.fs.staging + "/")[1].split("/", 3)[3]
                    self.references.append(rel_path)
            except Exception:
                self.logger.warn("Could not resolve path to a referenced calculation within the upload.")

        finally:
            os.chdir(cwd)

        if set_of_forces:
            phonopy_obj.set_forces(set_of_forces)
            phonopy_obj.produce_force_constants()

        self._phonopy_obj = phonopy_obj

    def parse_bandstructure(self):
        freqs, bands, bands_labels = self.properties.get_bandstructure()
        if freqs is None:
            return

        # convert THz to eV
        freqs = freqs * THzToEv

        # convert eV to J
        freqs = (freqs * ureg.eV).to('joules').magnitude

        sec_scc = self.archive.section_run[0].section_single_configuration_calculation[0]

        sec_k_band = sec_scc.m_create(KBand)
        sec_k_band.band_structure_kind = 'vibrational'

        for i in range(len(freqs)):
            freq = np.expand_dims(freqs[i], axis=0)
            sec_k_band_segment = sec_k_band.m_create(KBandSegment)
            sec_k_band_segment.band_energies = freq
            sec_k_band_segment.band_k_points = bands[i]
            sec_k_band_segment.band_segm_labels = bands_labels[i]

    def parse_dos(self):
        f, dos = self.properties.get_dos()

        # To match the shape given in meta data another dimension is added to the
        # array (spin degress of fredom is 1)
        dos = np.expand_dims(dos, axis=0)

        # convert THz to eV to Joules
        f = f * THzToEv
        f = (f * ureg.eV).to('joules').magnitude

        sec_scc = self.archive.section_run[0].section_single_configuration_calculation[0]
        sec_dos = sec_scc.m_create(Dos)
        sec_dos.dos_kind = 'vibrational'
        sec_dos.dos_values = dos
        sec_dos.dos_energies = f

    def parse_thermodynamical_properties(self):
        T, fe, _, cv = self.properties.get_thermodynamical_properties()

        n_atoms = len(self.phonopy_obj.unitcell)
        n_atoms_supercell = len(self.phonopy_obj.supercell)

        fe = fe / n_atoms

        # The thermodynamic properties are reported by phonopy for the base
        # system. Since the values in the metainfo are stored per the referenced
        # system, we need to multiple by the size factor between the base system
        # and the supersystem used in the calculations.
        cv = cv * (n_atoms_supercell / n_atoms)

        # convert to SI units
        fe = (fe * ureg.eV).to('joules').magnitude

        cv = (cv * ureg.eV / ureg.K).to('joules/K').magnitude

        sec_run = self.archive.section_run[0]
        sec_scc = sec_run.section_single_configuration_calculation

        sec_frame_sequence = sec_run.m_create(FrameSequence)
        sec_frame_sequence.frame_sequence_local_frames_ref = sec_scc

        sec_thermo_prop = sec_frame_sequence.m_create(ThermodynamicalProperties)
        sec_thermo_prop.thermodynamical_property_temperature = T
        sec_thermo_prop.vibrational_free_energy_at_constant_volume = fe
        sec_thermo_prop.thermodynamical_property_heat_capacity_C_v = cv

        sec_sampling_method = sec_run.m_create(SamplingMethod)
        sec_sampling_method.sampling_method = 'taylor_expansion'
        sec_sampling_method.sampling_method_expansion_order = 2

        sec_frame_sequence.frame_sequence_to_sampling_ref = sec_sampling_method

    def parse_ref(self):
        sec_scc = self.archive.section_run[0].section_single_configuration_calculation[0]
        for ref in self.references:
            sec_calc_refs = sec_scc.m_create(CalculationToCalculationRefs)
            sec_calc_refs.calculation_to_calculation_kind = 'source_calculation'
            sec_calc_refs.calculation_to_calculation_external_url = ref

    def parse(self, filepath, archive, logger, **kwargs):
        self.mainfile = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging
        self._kwargs.update(kwargs)

        self._metainfo_env = m_env

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'Phonopy'
        sec_run.program_version = phonopy.__version__

        phonopy_obj = self.phonopy_obj

        pbc = np.array((1, 1, 1), bool)

        unit_cell = self.phonopy_obj.unitcell.get_cell()
        unit_pos = self.phonopy_obj.unitcell.get_positions()
        unit_sym = np.array(self.phonopy_obj.unitcell.get_chemical_symbols())

        super_cell = self.phonopy_obj.supercell.get_cell()
        super_pos = self.phonopy_obj.supercell.get_positions()
        super_sym = np.array(self.phonopy_obj.supercell.get_chemical_symbols())

        unit_cell = (unit_cell * ureg.angstrom).to('meter').magnitude
        unit_pos = (unit_pos * ureg.angstrom).to('meter').magnitude

        super_cell = (super_cell * ureg.angstrom).to('meter').magnitude
        super_pos = (super_pos * ureg.angstrom).to('meter').magnitude

        try:
            displacement = np.linalg.norm(phonopy_obj.displacements[0][1:])
            displacement = (displacement * ureg.angstrom).to('meter').magnitude
        except Exception:
            displacement = None

        supercell_matrix = phonopy_obj.supercell_matrix
        sym_tol = phonopy_obj.symmetry.tolerance

        sec_system_unit = sec_run.m_create(System)
        sec_system_unit.configuration_periodic_dimensions = pbc
        sec_system_unit.atom_labels = unit_sym
        sec_system_unit.atom_positions = unit_pos
        sec_system_unit.simulation_cell = unit_cell

        sec_system = sec_run.m_create(System)
        sec_system_to_system_refs = sec_system.m_create(SystemToSystemRefs)
        sec_system_to_system_refs.system_to_system_kind = 'subsystem'
        sec_system_to_system_refs.system_to_system_ref = sec_system_unit
        sec_system.configuration_periodic_dimensions = pbc
        sec_system.atom_labels = super_sym
        sec_system.atom_positions = super_pos
        sec_system.simulation_cell = super_cell
        sec_system.SC_matrix = supercell_matrix
        sec_system.x_phonopy_original_system_ref = sec_system_unit

        sec_method = sec_run.m_create(Method)
        # TODO I put this so as to have a recognizable section method, but metainfo
        # should be expanded to include phonon related method parameters
        sec_method.electronic_structure_method = 'DFT'
        sec_method.x_phonopy_symprec = sym_tol
        if displacement is not None:
            sec_method.x_phonopy_displacement = displacement

        try:
            force_constants = phonopy_obj.get_force_constants()
            force_constants = (force_constants * ureg.eV / ureg.angstrom ** 2).to('J/(m**2)').magnitude
        except Exception:
            self.logger.error('Error producing force constants.')
            return

        sec_scc = sec_run.m_create(SingleConfigurationCalculation)
        sec_scc.single_configuration_calculation_to_system_ref = sec_system
        sec_scc.single_configuration_to_calculation_method_ref = sec_method
        sec_scc.hessian_matrix = force_constants

        # get bandstructure configuration file
        maindir = os.path.dirname(self.mainfile)
        files = [f for f in os.listdir(maindir) if f.endswith('.conf')]
        self._kwargs.update({'band_conf': os.path.join(maindir, files[0]) if files else None})
        self.properties = PhononProperties(self.phonopy_obj, self.logger, **self._kwargs)

        self.parse_bandstructure()
        self.parse_dos()
        self.parse_thermodynamical_properties()
        self.parse_ref()

        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.workflow_type = 'phonon'
        sec_phonon = sec_workflow.m_create(Phonon)
        sec_phonon.force_calculator = self.calculator
        vol = np.dot(unit_cell[0], np.cross(unit_cell[1], unit_cell[2]))
        sec_phonon.mesh_density = np.prod(self.properties.mesh) / vol
        n_imaginary = np.count_nonzero(self.properties.frequencies < 0)
        sec_phonon.n_imaginary_frequencies = n_imaginary
        if phonopy_obj.nac_params:
            sec_phonon.with_non_analytic_correction = True




#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import typing
import click
import datetime
import elasticsearch_dsl
import elasticsearch
import sys
import threading

from nomad import processing as proc, search, infrastructure, utils, config
from nomad.cli.cli import cli


def __run_parallel(
        uploads, parallel: int, callable, label: str):
    if isinstance(uploads, (tuple, list)):
        uploads_count = len(uploads)

    else:
        uploads_count = uploads.count()
        uploads = list(uploads)  # copy the whole mongo query set to avoid cursor timeouts

    cv = threading.Condition()
    threads: typing.List[threading.Thread] = []

    state = dict(
        completed_count=0,
        skipped_count=0,
        available_threads_count=parallel)

    logger = utils.get_logger(__name__)

    print('%d uploads selected, %s ...' % (uploads_count, label))

    def process_upload(upload: proc.Upload):
        logger.info('%s started' % label, upload_id=upload.upload_id)

        completed = False
        if callable(upload, logger):
            completed = True

        with cv:
            state['completed_count'] += 1 if completed else 0
            state['skipped_count'] += 1 if not completed else 0
            state['available_threads_count'] += 1

            print(
                '   %s %s and skipped %s of %s uploads' %
                (label, state['completed_count'], state['skipped_count'], uploads_count))

            cv.notify()

    for upload in uploads:
        logger.info(
            'cli schedules parallel %s processing for upload' % label,
            current_process=upload.current_process,
            current_task=upload.current_task, upload_id=upload.upload_id)
        with cv:
            cv.wait_for(lambda: state['available_threads_count'] > 0)
            state['available_threads_count'] -= 1
            thread = threading.Thread(target=lambda: process_upload(upload))
            threads.append(thread)
            thread.start()

    for thread in threads:
        thread.join()


def __run_processing(
        uploads, parallel: int, process, label: str, reprocess_running: bool = False,
        wait_for_tasks: bool = True, reset_first: bool = False):

    def run_process(upload, logger):
        logger.info(
            'cli calls %s processing' % label,
            current_process=upload.current_process,
            current_task=upload.current_task, upload_id=upload.upload_id)
        if upload.process_running and not reprocess_running:
            logger.warn(
                'cannot trigger %s, since the upload is already/still processing' % label,
                current_process=upload.current_process,
                current_task=upload.current_task, upload_id=upload.upload_id)
            return False

        if reset_first:
            upload.reset(force=True)
        elif upload.process_running:
            tasks_status = upload.tasks_status
            if tasks_status == proc.RUNNING:
                tasks_status = proc.FAILURE
            upload.reset(force=True, tasks_status=tasks_status)

        process(upload)
        if wait_for_tasks:
            upload.block_until_complete(interval=.5)
        else:
            upload.block_until_process_complete(interval=.5)

        if upload.tasks_status == proc.FAILURE:
            logger.info('%s with failure' % label, upload_id=upload.upload_id)

        logger.info('%s complete' % label, upload_id=upload.upload_id)
        return True

    __run_parallel(uploads, parallel=parallel, callable=run_process, label=label)


@cli.group(help='''The nomad admin commands to do nasty stuff directly on the databases.
                     Remember: With great power comes great responsibility!''')
@click.pass_context
def admin(ctx):
    pass


@admin.command(help='Reset/remove all databases.')
@click.option('--remove', is_flag=True, help='Do not just reset all dbs, but also remove them.')
@click.option('--i-am-really-sure', is_flag=True, help='Must be set for the command to to anything.')
def reset(remove, i_am_really_sure):
    if not i_am_really_sure:
        print('You do not seem to be really sure about what you are doing.')
        sys.exit(1)

    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    infrastructure.reset(remove)


@admin.command(help='Reset all "stuck" in processing uploads and calc in low level mongodb operations.')
@click.option('--zero-complete-time', is_flag=True, help='Sets the complete time to epoch zero.')
def reset_processing(zero_complete_time):
    infrastructure.setup_mongo()

    def reset_collection(cls):
        in_processing = cls.objects(process_status__in=[proc.PROCESS_RUNNING, proc.base.PROCESS_CALLED])
        print('%d %s processes need to be reset due to incomplete process' % (in_processing.count(), cls.__name__))
        in_processing.update(
            process_status=None,
            current_process=None,
            worker_hostname=None,
            celery_task_id=None,
            errors=[], warnings=[],
            complete_time=datetime.datetime.fromtimestamp(0) if zero_complete_time else datetime.datetime.now(),
            current_task=None,
            tasks_status=proc.base.CREATED)

        in_tasks = cls.objects(tasks_status__in=[proc.PENDING, proc.RUNNING])
        print('%d %s processes need to be reset due to incomplete tasks' % (in_tasks.count(), cls.__name__))
        in_tasks.update(
            current_task=None,
            tasks_status=proc.base.CREATED,
            errors=[], warnings=[],
            complete_time=datetime.datetime.fromtimestamp(0) if zero_complete_time else datetime.datetime.now())

    reset_collection(proc.Calc)
    reset_collection(proc.Upload)


@admin.command(help='Check and lift embargo of data with expired embargo period.')
@click.option('--dry', is_flag=True, help='Do not lift the embargo, just show what needs to be done.')
@click.option('--parallel', default=1, type=int, help='Use the given amount of parallel processes. Default is 1.')
def lift_embargo(dry, parallel):
    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    request = search.SearchRequest()
    request.q = elasticsearch_dsl.Q('term', with_embargo=True) & elasticsearch_dsl.Q('term', published=True)
    request.quantity('upload_id', 1000)
    result = request.execute()

    uploads_to_repack = []
    for upload_id in result['quantities']['upload_id']['values']:
        upload = proc.Upload.get(upload_id)
        embargo_length = upload.embargo_length
        if embargo_length is None:
            embargo_length = 36
            upload.embargo_length = 36

        if upload.upload_time + datetime.timedelta(days=int(embargo_length * 365 / 12)) < datetime.datetime.now():
            print('need to lift the embargo of %s (upload_time=%s, embargo=%d)' % (
                upload.upload_id, upload.upload_time, embargo_length))

            if not dry:
                proc.Calc._get_collection().update_many(
                    {'upload_id': upload_id},
                    {'$set': {'metadata.with_embargo': False}})
                uploads_to_repack.append(upload)
                upload.save()

                with upload.entries_metadata() as entries:
                    search.index_all(entries)

    if not dry:
        __run_processing(
            uploads_to_repack, parallel, lambda upload: upload.re_pack(), 're-packing',
            wait_for_tasks=False)


@admin.command()
@click.option('--threads', type=int, default=1, help='Number of threads to use.')
@click.option('--code', multiple=True, type=str, help='Index only calculcations of given codes.')
@click.option('--dry', is_flag=True, help='Do not index, just compute entries.')
@click.option('--in-place', is_flag=True, default=False, help='Perform indexing in the current elastic search index. Meant only for small reindex operations.')
@click.option('-n', type=int, default=None, help='Number of calculations to process. Leave undefined to process all calculations.')
@click.option('--source',
              type=click.Choice(['mongo', 'es'], case_sensitive=True))
def index_materials(threads, code, dry, in_place, n, source):
    """(Re-)index all materials.

    This command will is used to completely rebuild the materials index or
    parts of it. You can choose which source is used for building the index: In
    general ElasticSearch is preferred, but if it is not available also MongoDB
    can be used.

    By default this command will start to rebuild the index from scratch.  This
    can be done on a "live" system because a new temporary index is used. If
    you use the --in-place option, the indexing will be run on the same index
    that is currently in use.
    """

    from nomad.datamodel.material import Material, Calculation
    from nomad.datamodel.encyclopedia import EncyclopediaMetadata
    from nomad.search import material_document
    from nomad.datamodel.material import Material, Calculation, Method, Properties, IdealizedStructure, Energies, Workflow, Bulk

    def create_entry(material, calc, in_place):
        """Creates an ES update operation that inserts the full material info
        if entry does not exists, otherwise only adds the calculation into the
        nested subdocument, possibly replacing the old one if in_place is True.
        """
        entry = {}
        entry['_op_type'] = 'update'
        entry['_index'] = target_index_name
        entry['_id'] = material.material_id
        entry['_type'] = 'doc'
        entry['_source'] = {
            "upsert": material.m_to_dict(include_defaults=False, partial="es"),
            "doc_as_upsert": False,
            "script": {
                "params": {
                    "calc": calc.m_to_dict(include_defaults=False, partial="es")
                },
            }
        }
        if in_place:
            entry['_source']["script"]["source"] = "ctx._source.calculations.removeIf(x -> x.calc_id == params.calc.calc_id); ctx._source.calculations.add(params.calc)"
        else:
            entry['_source']["script"]["source"] = "ctx._source.calculations.add(params.calc)"
        return entry

    chunk_size = 500
    infrastructure.setup_mongo()
    client = infrastructure.setup_elastic()

    # In order to do the reindexing with zero downtime, two different indices
    # are rotated and an alias is used
    old_index_name = list(client.indices.get(config.elastic.materials_index_name).keys())[0]
    if in_place:
        target_index_name = old_index_name
    else:
        if old_index_name == config.elastic.materials_index_name + "_a":
            target_index_name = config.elastic.materials_index_name + "_b"
        elif old_index_name == config.elastic.materials_index_name + "_b":
            target_index_name = config.elastic.materials_index_name + "_a"
        else:
            raise ValueError(
                "Unrecognized index name accociated with the alias {}"
                .format(config.elastic.materials_index_name)
            )

    if source == "mongo":
        mongo_db = infrastructure.mongo_client[config.mongo.db_name]
        mongo_collection = mongo_db['archive']
        if code:
            collection = mongo_collection.find({"section_metadata.dft.code_name": {"$in": code}})
        else:
            collection = mongo_collection.find()
        all_calcs = collection.count()
        print('indexing materials from %d calculations ...' % all_calcs)

        # Bulk update
        def elastic_updates():
            with utils.ETA(all_calcs, '   index %10d of %10d calcs, ETA %s') as eta:
                i_calc = 0
                for mongo_archive in collection:
                    i_calc += 1
                    if n is not None:
                        if i_calc > n:
                            return
                    eta.add()

                    # Do not process entries that do not have the material
                    # information
                    try:
                        status = mongo_archive["section_metadata"]["encyclopedia"]["status"]
                        if status != EncyclopediaMetadata.status.type.success:
                            raise AttributeError
                    except (KeyError, AttributeError, IndexError):
                        continue

                    # Create material information
                    metadata = mongo_archive["section_metadata"]
                    encyclopedia = EncyclopediaMetadata.m_from_dict(metadata["encyclopedia"])
                    dft = metadata["dft"]
                    material: Material = Material()
                    material.material_id = encyclopedia.material.material_id
                    material.material_type = encyclopedia.material.material_type
                    material.material_name = encyclopedia.material.material_name
                    material.material_classification = encyclopedia.material.material_classification
                    material.formula = encyclopedia.material.formula
                    material.formula_reduced = encyclopedia.material.formula_reduced
                    material.species_and_counts = encyclopedia.material.species_and_counts
                    material.species = encyclopedia.material.species
                    enc_bulk = encyclopedia.material.bulk
                    if enc_bulk:
                        bulk = Bulk.m_from_dict(enc_bulk.m_to_dict())
                        material.m_add_sub_section(Material.bulk, bulk)

                    # Create calculation info for this entry
                    calc = Calculation()
                    calc.calc_id = metadata["calc_id"]
                    calc.upload_id = metadata["upload_id"]
                    mongo_calc = proc.Calc.get(calc.calc_id)
                    calc.published = mongo_calc["metadata"]["published"]
                    calc.with_embargo = mongo_calc["metadata"]["with_embargo"]
                    calc.owners = [mongo_calc["metadata"]["uploader"]] + mongo_calc["metadata"]["shared_with"]
                    enc_idealized_structure = encyclopedia.material.idealized_structure
                    idealized_structure = IdealizedStructure()
                    cell_volume = enc_idealized_structure.cell_volume
                    if cell_volume is not None:
                        idealized_structure.cell_volume = cell_volume
                    idealized_structure.lattice_parameters = enc_idealized_structure.lattice_parameters
                    calc.m_add_sub_section(Calculation.idealized_structure, idealized_structure)
                    enc_method = encyclopedia.method
                    method = Method.m_from_dict(enc_method.m_to_dict())
                    method.program_name = dft["code_name"]
                    method.program_version = dft["code_version"]
                    method.basis_set = dft["basis_set"]
                    calc.m_add_sub_section(Calculation.method, method)
                    enc_props = encyclopedia.properties

                    # Properties may not exist at all
                    if enc_props is not None:
                        properties = Properties()

                        # Energies may not be present in all calculations
                        try:
                            energies = Energies.m_from_dict(enc_props.energies.m_to_dict())
                            properties.m_add_sub_section(Properties.energies, energies)
                        except AttributeError:
                            pass

                        properties.has_electronic_dos = enc_props.electronic_dos is not None
                        properties.has_electronic_band_structure = enc_props.electronic_band_structure is not None
                        properties.has_thermodynamical_properties = enc_props.thermodynamical_properties is not None
                        atomic_density = enc_props.atomic_density
                        if atomic_density is not None:
                            properties.atomic_density = atomic_density
                        mass_density = enc_props.mass_density
                        if mass_density is not None:
                            properties.mass_density = mass_density
                        band_gap = enc_props.band_gap
                        if band_gap is not None:
                            properties.band_gap = band_gap
                        band_gap_direct = enc_props.band_gap_direct
                        if band_gap_direct is not None:
                            properties.band_gap_direct = band_gap_direct
                        calc.m_add_sub_section(Calculation.properties, properties)

                    workflow = Workflow()
                    workflow.workflow_type = encyclopedia.calculation.calculation_type
                    calc.m_add_sub_section(Calculation.workflow, workflow)
                    material.m_add_sub_section(Material.calculations, calc)

                    # Update entry that inserts the full material info if entry
                    # does not exists, otherwise only adds the calculation into
                    # the nested subdocument
                    yield create_entry(material, calc, in_place)

    elif source == "es":
        s = elasticsearch_dsl.Search(index=config.elastic.index_name)
        filters = [elasticsearch_dsl.Q("term", encyclopedia__status="success")]
        if code:
            filters.append(elasticsearch_dsl.Q("terms", dft__code_name=code))
        query = elasticsearch_dsl.Q(
            "bool",
            filter=filters,
        )
        s = s.query(query)
        s = s.extra(**{
            "size": 0,
        })
        all_calcs = s.execute().hits.total
        print('indexing materials from %d calculations ...' % all_calcs)

        def elastic_updates():
            with utils.ETA(all_calcs, '   index %10d of %10d calcs, ETA %s', chunk_size) as eta:

                s = elasticsearch_dsl.Search(index=config.elastic.index_name)
                filters = [elasticsearch_dsl.Q("term", encyclopedia__status="success")]
                if code:
                    filters.append(elasticsearch_dsl.Q("terms", dft__code_name=code))
                query = elasticsearch_dsl.Q(
                    "bool",
                    filter=filters,
                )
                s = s.query(query)
                s = s.extra(**{
                    "size": chunk_size,
                })
                i_calc = 0
                for hit in s.scan():
                    i_calc += 1
                    if n is not None:
                        if i_calc > n:
                            return
                    eta.add()

                    material: Material = Material()
                    calc = Calculation()

                    # Check that all required information exists. If not, the
                    # calculation is skipped.
                    try:
                        material.material_id = hit.encyclopedia.material.material_id
                        material.material_type = hit.encyclopedia.material.material_type
                        material.formula = hit.encyclopedia.material.formula
                        material.formula_reduced = hit.encyclopedia.material.formula_reduced
                        material.species_and_counts = hit.encyclopedia.material.species_and_counts
                        material.species = hit.encyclopedia.material.species
                        calc.calc_id = hit.calc_id
                        calc.upload_id = hit.upload_id
                        calc.published = hit.published
                        calc.with_embargo = hit.with_embargo
                        calc.owners = [x.user_id for x in hit.owners]
                        idealized_structure = IdealizedStructure.m_from_dict(hit.encyclopedia.material.idealized_structure.to_dict())
                        calc.m_add_sub_section(Calculation.idealized_structure, idealized_structure)

                        method = Method.m_from_dict(hit.encyclopedia.method.to_dict())
                        method.program_name = hit.dft.code_name
                        method.program_version = hit.dft.code_version
                        method.basis_set = hit.dft.basis_set
                        calc.m_add_sub_section(Calculation.method, method)

                        workflow = Workflow()
                        workflow.workflow_type = hit.encyclopedia.calculation.calculation_type
                        calc.m_add_sub_section(Calculation.workflow, workflow)
                    except AttributeError:
                        continue

                    # Not all materials have a name
                    try:
                        material.material_name = hit.encyclopedia.material.material_name
                    except AttributeError:
                        pass

                    # Not all materials have a bulk section
                    try:
                        bulk = Bulk.m_from_dict(hit.encyclopedia.material.bulk)
                        material.m_add_sub_section(Material.bulk, bulk)
                    except AttributeError:
                        pass

                    # Properties may not exist at all
                    try:
                        enc_properties = hit.encyclopedia.properties
                    except AttributeError:
                        pass
                    else:
                        properties = Properties()

                        # Energies may not be present in all calculations
                        try:
                            energies = Energies.m_from_dict(enc_properties.energies.to_dict())
                            properties.m_add_sub_section(Properties.energies, energies)
                        except AttributeError:
                            pass

                        # Gather the boolean flags that indicate the presence of
                        # certain properties
                        try:
                            properties.has_electronic_dos = enc_properties.electronic_dos is not None
                        except AttributeError:
                            properties.has_electronic_dos = False
                        try:
                            properties.has_electronic_band_structure = enc_properties.electronic_band_structure is not None
                        except AttributeError:
                            properties.has_electronic_band_structure = False
                        try:
                            properties.has_thermodynamical_properties = enc_properties.thermodynamical_properties is not None
                        except AttributeError:
                            properties.has_thermodynamical_properties = False

                        # Not all materials have an atomic density
                        try:
                            properties.atomic_density = enc_properties.atomic_density
                        except AttributeError:
                            pass

                        # Not all materials have a mass density
                        try:
                            properties.mass_density = enc_properties.mass_density
                        except AttributeError:
                            pass

                        # Not all materials have band gaps
                        try:
                            properties.band_gap = enc_properties.band_gap
                        except AttributeError:
                            pass

                        # Not all materials have band gap type
                        try:
                            properties.band_gap_direct = enc_properties.band_gap_direct
                        except AttributeError:
                            pass

                        calc.m_add_sub_section(Calculation.properties, properties)

                    material.m_add_sub_section(Material.calculations, calc)

                    yield create_entry(material, calc, in_place)

    if dry:
        for _ in elastic_updates():
            pass
    else:
        # Create new index into which the data will be inserted. The old index will
        # keep working while the new index is being built
        material_document.init(index=target_index_name)

        if threads > 1:
            print('  use %d threads' % threads)
            for _ in elasticsearch.helpers.parallel_bulk(
                    infrastructure.elastic_client, elastic_updates(), chunk_size=chunk_size,
                    thread_count=threads):
                pass
        else:
            elasticsearch.helpers.bulk(
                infrastructure.elastic_client, elastic_updates())
            search.refresh()

        # Changes materials index alias to point to the new index and remove the
        # old index.
        if not in_place:
            new_index = elasticsearch_dsl.Index(target_index_name)
            new_index.put_alias(name=config.elastic.materials_index_name)
            old_index = elasticsearch_dsl.Index(old_index_name)
            old_index.delete()

    print('')
    print('indexing completed')


@admin.group(help='Generate scripts and commands for nomad operation.')
def ops():
    pass


@ops.group(help='Tools for managing the DOS similarity data.')
def similarity():
    pass


@ops.command(help=('Dump the mongo (calculation metadata) db.'))
@click.option('--restore', is_flag=True, help='Do not dump, but restore.')
def dump(restore: bool):
    date_str = datetime.datetime.utcnow().strftime('%Y_%m_%d')
    print('mongodump --host {} --port {} --db {} -o /backup/fairdi/mongo/{}'.format(
        config.mongo.host, config.mongo.port, config.mongo.db_name, date_str))


@ops.command(help=('Restore the mongo (calculation metadata) db.'))
@click.argument('PATH_TO_DUMP', type=str, nargs=1)
def restore(path_to_dump):
    print('mongorestore --host {} --port {} --db {} {}'.format(
        config.mongo.host, config.mongo.port, config.mongo.db_name, path_to_dump))


@ops.command(help=('Generate an nginx.conf to serve the GUI and proxy pass to API container.'))
@click.option('--prefix', type=str, default=config.services.api_base_path, help='Alter the url path prefix.')
@click.option('--host', type=str, default=config.services.api_host, help='Alter the NOMAD app host.')
@click.option('--port', type=str, default=config.services.api_port, help='Alter the NOMAD port host.')
@click.option('--server/--no-server', default=True, help='Control writing of the outer server {} block. '
              'Useful when conf file is included within another nginx.conf.')
def nginx_conf(prefix, host, port, server):
    prefix = prefix.rstrip('/')
    prefix = '/%s' % prefix.lstrip('/')

    if server:
        print('''server {
    listen        80;
    server_name   www.example.com;
    proxy_set_header Host $host;
        ''')

    print('''
    location / {{
        proxy_pass http://{1}:{2};
    }}

    location ~ {0}\\/?(gui)?$ {{
        rewrite ^ {0}/gui/ permanent;
    }}

    location {0}/gui/ {{
        proxy_intercept_errors on;
        error_page 404 = @redirect_to_index;
        proxy_pass http://{1}:{2};
    }}

    location @redirect_to_index {{
        rewrite ^ {0}/gui/index.html break;
        proxy_pass http://{1}:{2};
    }}

    location ~ \\/gui\\/(service-worker\\.js|meta\\.json)$ {{
        add_header Last-Modified $date_gmt;
        add_header Cache-Control 'no-store, no-cache, must-revalidate, proxy-revalidate, max-age=0';
        if_modified_since off;
        expires off;
        etag off;
        proxy_pass http://{1}:{2};
    }}

    location ~ \\/api\\/uploads\\/?$ {{
        client_max_body_size 35g;
        proxy_request_buffering off;
        proxy_pass http://{1}:{2};
    }}

    location ~ \\/api\\/(raw|archive) {{
        proxy_buffering off;
        proxy_pass http://{1}:{2};
    }}

    location ~ \\/api\\/mirror {{
        proxy_buffering off;
        proxy_read_timeout 600;
        proxy_pass http://{1}:{2};
    }}

    location ~ \\/encyclopedia\\/ {{
        proxy_intercept_errors on;
        error_page 404 = @redirect_to_encyclopedia_index;
        proxy_pass http://{1}:{2};
    }}

    location @redirect_to_encyclopedia_index {{
        rewrite ^ {0}/encyclopedia/index.html break;
        proxy_pass http://{1}:{2};
    }}
'''.format(prefix, host, port))
    if server:
        print('}')


@ops.command(help='Updates the AFLOW prototype information using the latest online version and writes the results to a python module in the given FILEPATH.')
@click.argument('FILEPATH', nargs=1, type=str)
@click.option('--matches-only', is_flag=True, help='Only update the match information that depends on the symmetry analysis settings. Will not perform and online update.')
@click.pass_context
def prototypes_update(ctx, filepath, matches_only):
    from nomad.cli.admin import prototypes
    prototypes.update_prototypes(ctx, filepath, matches_only)


@ops.command(help='Updates the springer database in nomad.config.normalize.springer_db_path.')
@click.option('--max-n-query', default=10, type=int, help='Number of unsuccessful springer request before returning an error. Default is 10.')
@click.option('--retry-time', default=120, type=int, help='Time in seconds to retry after unsuccessful request. Default is 120.')
def springer_update(max_n_query, retry_time):
    from nomad.cli.admin import springer
    springer.update_springer(max_n_query, retry_time)


@similarity.command(help='Updates the msgpack file containing the similarity information.')
@click.option('--dir', "-d", "input_dir", type=str, help='Path of the folder containing the raw similarity information files')
@click.option('--out', "-o", type=str, help='Path of the output msgpack file.')
@click.option('--verbose', is_flag=True, help='Enable verbose output.')
def update(input_dir, out, verbose):
    from nomad.cli.admin import similarity
    similarity.update(input_dir, out, verbose)


@similarity.command(help='Ingests the given similarity information from an msgpack file into MongoDB.')
@click.option('--in', "-i", "input_path", type=str, help='Path of the ingested msgpack file.')
@click.option('--batch_size', type=int, default=10000, help='Batch size for MongoDB bulk ingestion.')
@click.option('--verbose', is_flag=True, help='Enable verbose output.')
def ingest(input_path, batch_size, verbose):
    from nomad.cli.admin import similarity
    similarity.ingest(input_path, batch_size, verbose)


@ops.command(help='Configures the GUIs based on NOMAD config.')
def gui_config():
    import os
    import os.path
    from nomad import config
    import glob
    import shutil
    import json

    gui_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../app/flask/static/gui'))
    run_gui_folder = os.path.join(gui_folder, '../.gui_configured')

    # copy
    shutil.rmtree(run_gui_folder, ignore_errors=True)
    shutil.copytree(gui_folder, run_gui_folder)

    # setup the env
    env_js_file = os.path.join(run_gui_folder, 'env.js')
    if not os.path.exists(env_js_file):
        with open(env_js_file, 'wt') as f:
            f.write(f'''
window.nomadEnv = {{
    'appBase': '{config.services.api_base_path}',
    'keycloakBase': 'https://nomad-lab.eu/fairdi/keycloak/auth/',
    'keycloakRealm': '{config.keycloak.realm_name}',
    'keycloakClientId': '{config.keycloak.client_id}',
    'debug': false,
    'encyclopediaEnabled': {'true' if config.encyclopedia_enabled else 'false'},
    'aitoolkitEnabled': {'true' if config.aitoolkit_enabled else 'false'},
    'oasis': {'true' if config.keycloak.oasis else 'false'},
    'version': {json.dumps(config.meta.beta) if config.meta.beta else dict()}
}};''')

    # replace base path in all GUI files
    source_file_globs = [
        '**/*.json',
        '**/*.html',
        '**/*.js',
        '**/*.js.map',
        '**/*.css']
    for source_file_glob in source_file_globs:
        source_files = glob.glob(os.path.join(run_gui_folder, source_file_glob), recursive=True)
        for source_file in source_files:
            with open(source_file, 'rt') as f:
                file_data = f.read()
            file_data = file_data.replace('/fairdi/nomad/latest', config.services.api_base_path)
            with open(source_file, 'wt') as f:
                f.write(file_data)

    gui_folder = os.path.abspath(os.path.join(
        os.path.dirname(__file__), '../../app/flask/static/encyclopedia'))

    # setup the env
    conf_js_file = os.path.join(gui_folder, 'conf.js')
    if not os.path.exists(conf_js_file):
        with open(conf_js_file, 'wt') as f:
            f.write(('''
window.nomadEnv = {
    apiRoot: "%s/api/encyclopedia/",
    keycloakBase: "%s",
    keycloakRealm: "%s",
    keycloakClientId: "%s"
};''' % (
                config.services.api_base_path,
                config.keycloak.server_url,
                config.keycloak.realm_name,
                config.keycloak.client_id
            )))
